#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import copy

import numpy as np
import pandas as pd
import pints
import xarray as xr

import chi


class AveragedPredictiveModel(object):
    """
    A base class for predictive models whose parameters are drawn from
    a distribution.
    """

    def __init__(self, predictive_model):
        super(AveragedPredictiveModel, self).__init__()

        # Check inputs
        if not isinstance(predictive_model, chi.PredictiveModel):
            raise ValueError(
                'The provided predictive model has to be an instance of a '
                'chi.PredictiveModel.')

        self._predictive_model = predictive_model

    def get_dosing_regimen(self, final_time=None):
        """
        Returns the dosing regimen of the compound in form of a
        :class:`pandas.DataFrame`.

        The dataframe has a time, a duration, and a dose column, which indicate
        the time point and duration of the dose administration in the time
        units of the mechanistic model, :meth:`MechanisticModel.time_unit`. The
        dose column specifies the amount of the compound that is being
        administered in units of the drug amount variable of the mechanistic
        model.

        If an indefinitely administered dosing regimen is set, i.e. a
        finite duration and undefined number of doses, see
        :meth:`set_dosing_regimen`, only the first administration of the
        dose will appear in the dataframe. Alternatively, a final dose time
        ``final_time`` can be provided, up to which the dose events are
        registered.

        If no dosing regimen has been set, ``None`` is returned.

        Parameters
        ----------
        final_time
            Time up to which dose events are registered in the dataframe. If
            ``None``, all dose events are registered, except for indefinite
            dosing regimens. Here, only the first dose event is registered.
        """
        return self._predictive_model.get_dosing_regimen(final_time)

    def get_n_outputs(self):
        """
        Returns the number of outputs.
        """
        return self._predictive_model.get_n_outputs()

    def get_output_names(self):
        """
        Returns the output names.
        """
        return self._predictive_model.get_output_names()

    def get_predictive_model(self):
        """
        Returns the predictive model.
        """
        return self._predictive_model

    def sample(
            self, times, n_samples=None, seed=None, include_regimen=False):
        """
        Samples virtual measurements from the model of the data-generating
        process and returns them in form of a :class:`pandas.DataFrame`.
        """
        raise NotImplementedError

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen of the administered compound.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        .. note::
            This method requires a :class:`MechanisticModel` that supports
            dose administration.

        Parameters
        ----------
        dose
            The amount of the compound that is injected at each administration.
        start
            Start time of the treatment.
        duration
            Duration of dose administration. For a bolus injection, a dose
            duration of 1% of the time unit should suffice. By default the
            duration is set to 0.01 (bolus).
        period
            Periodicity at which doses are administered. If ``None`` the dose
            is administered only once.
        num
            Number of administered doses. If ``None`` and the periodicity of
            the administration is not ``None``, doses are administered
            indefinitely.
        """
        self._predictive_model.set_dosing_regimen(
            dose, start, duration, period, num)


class PosteriorPredictiveModel(AveragedPredictiveModel):
    r"""
    Implements the posterior predictive model of the modelled
    data-generating process and the associated parameter posterior
    distribution.

    A :class:`PosteriorPredictiveModel` is instantiated with a
    :class:`PredictiveModel` and samples from the associated posterior
    distribution in form of a :class:`xarray.Dataset`. The samples approximate
    the posterior distribution of the model parameters.

    Formally, the posterior predictive model is defined as

    .. math::
        p(y | \mathcal{D}, t) = \int \text{d}\psi \,
            p(y | \psi , t)\, p(\psi | \mathcal{D}),

    where :math:`p(y | \psi, t)` is the model of the data-generating process
    and :math:`p(\psi | \mathcal{D})` is the posterior distribution of
    the model parmeters.

    :param predictive_model: A predictive model which defines the distribution
        of observable biomarkers over time conditioned on parameter values.
    :type predictive_model: PredictiveModel
    :param posterior_samples: Samples from the posterior distribution of the
        model parameters.
    :type posterior_samples: xarray.Dataset
    :param param_map: A dictionary which can be used to map predictive model
        parameter names to the parameter names in the :class:`xarray.Dataset`.
        If ``None``, it is assumed that the names are identical.
    :type param_map: dict, optional
    """
    def __init__(
            self, predictive_model, posterior_samples, param_map=None):
        super(PosteriorPredictiveModel, self).__init__(predictive_model)

        # Check input
        if not isinstance(posterior_samples, xr.Dataset):
            raise TypeError(
                'The posterior samples have to be a xarray.Dataset.')

        dims = sorted(list(posterior_samples.dims))
        expected_dims = ['chain', 'draw', 'individual']
        if (len(dims) == 2):
            expected_dims = ['chain', 'draw']
        for dim in expected_dims:
            if dim not in dims:
                raise ValueError(
                    'The posterior samples must have the dimensions '
                    '(chain, draw, individual). The current dimensions are <'
                    + str(dims) + '>.')

        # Set default parameter map (no mapping)
        if param_map is None:
            param_map = {}

        try:
            param_map = dict(param_map)
        except (TypeError, ValueError):
            raise ValueError(
                'The parameter map has to be convertable to a python '
                'dictionary.')

        # Check that the parameters of the posterior can be identified in the
        # dataset
        self._parameter_names = self._check_parameters(
            posterior_samples, param_map)

        # Store posterior
        self._posterior = posterior_samples

    def _check_parameters(self, posterior_samples, param_map):
        """
        Checks whether the parameters of the posterior exist in the dataset
        and returns them.
        """
        # Create map from posterior parameter names to model parameter names
        model_names = self._predictive_model.get_parameter_names()
        for param_id, name in enumerate(model_names):
            try:
                model_names[param_id] = param_map[name]
            except KeyError:
                # The name is not mapped
                pass

        # Check that model names are in the dataset
        for parameter in model_names:
            if parameter not in posterior_samples.data_vars:
                raise ValueError(
                    'The parameter <' + str(parameter) + '> cannot be found '
                    'in the posterior.')

        return model_names

    def sample(
            self, times, n_samples=None, individual=None, seed=None,
            include_regimen=False, covariates=None):
        """
        Samples virtual measurements from the posterior predictive
        model and returns them in form of a :class:`pandas.DataFrame`.

        For each of the ``n_samples`` a parameter set is drawn from the
        approximate posterior distribution. These paramaters are then used to
        sample from the predictive model.

        :param times: Times for the virtual "measurements".
        :type times: list, numpy.ndarray of shape (n,)
        :param n_samples: The number of virtual "measurements" that are
            performed at each time point. If ``None`` the biomarkers are
            measured only once at each time point.
        :type n_samples: int, optional
        :param individual: The ID of the modelled individual. If
            ``None``, either the first ID or the population is simulated.
        :type individual: str, optional
        :param seed: Seed for the pseudo-random number generator.
        :type seed: int or :class:`numpy.random.Generator`, optional
        :param include_regimen: A boolean flag which determines whether the
            dosing regimen information is included in the output. Only possible
            when ``return_df=True``.
        :type include_regimen: bool, optional
        :param covariates: Covariate values, specifying the sampled
            subpopulation.
        :type covariates: List, np.ndarray of shape ``(n_cov,)`` or
            ``(n_samples, n_cov)``, optional
        :rtype: :class:`pandas.DataFrame`
        """
        # Make sure n_samples is an integer
        if n_samples is None:
            n_samples = 1
        n_samples = int(n_samples)

        # Check individual for population model
        if isinstance(self._predictive_model, chi.PopulationPredictiveModel):
            individual = None
        # Check individual for individual predictive model
        else:
            try:
                ids = self._posterior.individual
            except AttributeError:  # pragma: no cover
                # Posterior only contains information about one individual.
                # Hack to make this work
                individual = 'some value'
                ids = ['some value']

            # Get default individual
            if individual is None:
                individual = str(ids.data[0])

            # Make sure individual exists
            if individual not in ids:
                raise ValueError(
                    'The individual <' + str(individual) + '> could not be '
                    'found in the ID column.')

        # Sort times
        times = np.sort(times)

        # Instantiate random number generator for sampling from the posterior
        rng = np.random.default_rng(seed=seed)

        # Sort parameters into numpy array for simplified sampling
        n_chains = len(self._posterior.chain)
        n_parameters = self._predictive_model.n_parameters()
        try:
            n_draws = len(self._posterior.sel(
                    individual=individual).dropna(dim='draw').draw)
        # Note: ValueError -> KeyError for xarray>=0.19.0
        except (ValueError, KeyError):
            n_draws = len(self._posterior.dropna(dim='draw').draw)
        posterior = np.empty(shape=(n_chains * n_draws, n_parameters))
        for param_id, parameter in enumerate(self._parameter_names):
            try:
                posterior[:, param_id] = self._posterior[parameter].sel(
                    individual=individual).dropna(dim='draw').values.flatten()
            # Note: ValueError -> KeyError for xarray>=0.19.0
            except (ValueError, KeyError):
                # If individual dimension does not exist, the parameter must
                # be a population parameter.
                posterior[:, param_id] = self._posterior[
                    parameter].dropna(dim='draw').values.flatten()

        # Create container for samples
        container = pd.DataFrame(
            columns=['ID', 'Time', 'Observable', 'Value'])

        # Get model outputs
        outputs = self._predictive_model.get_output_names()

        # Draw samples
        sample_ids = np.arange(start=1, stop=n_samples+1)
        for sample_id in sample_ids:
            # Sample parameter from posterior
            parameters = rng.choice(posterior)

            # Sample from predictive model
            sample = self._predictive_model.sample(
                parameters, times, n_samples, rng, return_df=False,
                covariates=covariates)

            # Append samples to dataframe
            for output_id, name in enumerate(outputs):
                container = pd.concat([container, pd.DataFrame({
                    'ID': sample_id,
                    'Time': times,
                    'Observable': name,
                    'Value': sample[output_id, :, 0]})])

        # Add dosing regimen, if set
        final_time = np.max(times)
        regimen = self.get_dosing_regimen(final_time)
        if (regimen is not None) and (include_regimen is True):
            # Append dosing regimen only once for all samples
            container = pd.concat([container, regimen])

        return container


class PredictiveModel(object):
    r"""
    Implements a model of a data-generating process.

    The model is defined by an instance of a :class:`MechanisticModel` and an
    instance of an :class:`ErrorModel` for each mechanistic model output

    .. math::
        p(y | \psi , t),

    where :math:`y` are measurements of quantities of interest, :math:`\psi`
    are the model parameters and :math:`t` is the time.

    Parameters
    ----------
    mechanistic_model
        An instance of a :class:`MechanisticModel`.
    error_models
        A list of :class:`ErrorModel` instances, one for each model output of
        the mechanistic model.
    outputs
        A list of the model outputs, which maps the error models to the model
        outputs. If ``None`` the error models are assumed to be listed in the
        same order as the model outputs.
    """

    def __init__(self, mechanistic_model, error_models, outputs=None):
        super(PredictiveModel, self).__init__()

        # Check inputs
        if not isinstance(
                mechanistic_model, chi.MechanisticModel):
            raise TypeError(
                'The mechanistic model has to be an instance of a '
                'chi.MechanisticModel.')

        error_models = \
            error_models if isinstance(error_models, list) else [error_models]
        for error_model in error_models:
            if not isinstance(
                    error_model, (chi.ErrorModel, chi.ReducedErrorModel)):
                raise TypeError(
                    'All error models have to be instances of a '
                    'chi.ErrorModel.')

        # Copy mechanistic model
        mechanistic_model = mechanistic_model.copy()

        # Set outputs
        if outputs is not None:
            mechanistic_model.set_outputs(outputs)

        # Get number of outputs
        n_outputs = mechanistic_model.n_outputs()

        if len(error_models) != n_outputs:
            raise ValueError(
                'Wrong number of error models. One error model has to be '
                'provided for each mechanistic error model.')

        # Copy error models
        error_models = [
            copy.deepcopy(error_model) for error_model in error_models]

        # Remember models
        self._mechanistic_model = mechanistic_model
        self._error_models = error_models

        # Set parameter names and number of parameters
        self._set_error_model_parameter_names()
        self._set_number_and_parameter_names()

    def _set_error_model_parameter_names(self):
        """
        Resets the error model parameter names and prepends the output name
        if more than one output exists.
        """
        # Reset error model parameter names to defaults
        for error_model in self._error_models:
            error_model.set_parameter_names(None)

        # Rename error model parameters, if more than one output
        n_outputs = self._mechanistic_model.n_outputs()
        if n_outputs > 1:
            # Get output names
            outputs = self._mechanistic_model.outputs()

            for output_id, error_model in enumerate(self._error_models):
                # Get original parameter names
                names = error_model.get_parameter_names()

                # Prepend output name
                output = outputs[output_id]
                names = [output + ' ' + name for name in names]

                # Set new parameter names
                error_model.set_parameter_names(names)

    def _set_number_and_parameter_names(self):
        """
        Sets the number and names of the free model parameters.
        """
        # Get mechanistic model parameters
        parameter_names = self._mechanistic_model.parameters()

        # Get error model parameters
        for error_model in self._error_models:
            parameter_names += error_model.get_parameter_names()

        # Update number and names
        self._parameter_names = parameter_names
        self._n_parameters = len(self._parameter_names)

    def fix_parameters(self, name_value_dict):
        """
        Fixes the value of model parameters, and effectively removes them as a
        parameter from the model. Fixing the value of a parameter at ``None``,
        sets the parameter free again.

        Parameters
        ----------
        name_value_dict
            A dictionary with model parameter names as keys, and parameter
            value as values.
        """
        # Check type of dictionanry
        try:
            name_value_dict = dict(name_value_dict)
        except (TypeError, ValueError):
            raise ValueError(
                'The name-value dictionary has to be convertable to a python '
                'dictionary.')

        # Get submodels
        mechanistic_model = self._mechanistic_model
        error_models = self._error_models

        # Convert models to reduced models
        if not isinstance(mechanistic_model, chi.ReducedMechanisticModel):
            mechanistic_model = chi.ReducedMechanisticModel(mechanistic_model)
        for model_id, error_model in enumerate(error_models):
            if not isinstance(error_model, chi.ReducedErrorModel):
                error_models[model_id] = chi.ReducedErrorModel(error_model)

        # Fix model parameters
        mechanistic_model.fix_parameters(name_value_dict)
        for error_model in error_models:
            error_model.fix_parameters(name_value_dict)

        # If no parameters are fixed, get original model back
        if mechanistic_model.n_fixed_parameters() == 0:
            mechanistic_model = mechanistic_model.mechanistic_model()

        for model_id, error_model in enumerate(error_models):
            if error_model.n_fixed_parameters() == 0:
                error_model = error_model.get_error_model()
                error_models[model_id] = error_model

        # Safe reduced models
        self._mechanistic_model = mechanistic_model
        self._error_models = error_models

        # Update names and number of parameters
        self._set_number_and_parameter_names()

    def get_dosing_regimen(self, final_time=None):
        """
        Returns the dosing regimen of the compound in form of a
        :class:`pandas.DataFrame`.

        The dataframe has a time, a duration, and a dose column, which indicate
        the time point and duration of the dose administration in the time
        units of the mechanistic model, :meth:`MechanisticModel.time_unit`. The
        dose column specifies the amount of the compound that is being
        administered in units of the drug amount variable of the mechanistic
        model.

        If an indefinitely administered dosing regimen is set, i.e. a
        finite duration and undefined number of doses, see
        :meth:`set_dosing_regimen`, only the first administration of the
        dose will appear in the dataframe. Alternatively, a final dose time
        ``final_time`` can be provided, up to which the dose events are
        registered.

        If no dosing regimen has been set, ``None`` is returned.

        Parameters
        ----------
        final_time
            Time up to which dose events are registered in the dataframe. If
            ``None``, all dose events are registered, except for indefinite
            dosing regimens. Here, only the first dose event is registered.
        """
        # Get regimen
        try:
            regimen = self._mechanistic_model.dosing_regimen()
        except AttributeError:
            # The model does not support dosing regimens
            return None

        # Return None if regimen is not set
        if regimen is None:
            return regimen

        # Make sure that final_time is positive
        if final_time is None:
            final_time = np.inf

        # Sort regimen into dataframe
        regimen_df = pd.DataFrame(columns=['Time', 'Duration', 'Dose'])
        for dose_event in regimen.events():
            # Get dose amount
            dose_rate = dose_event.level()
            dose_duration = dose_event.duration()
            dose_amount = dose_rate * dose_duration

            # Get dosing time points
            start_time = dose_event.start()
            period = dose_event.period()
            n_doses = dose_event.multiplier()

            if start_time > final_time:
                # Dose event exceeds final dose time and is therefore
                # not registered
                continue

            if period == 0:
                # Dose is administered only once
                regimen_df = pd.concat([regimen_df, pd.DataFrame({
                    'Time': [start_time],
                    'Duration': [dose_duration],
                    'Dose': [dose_amount]})])

                # Continue to next dose event
                continue

            if n_doses == 0:
                # The dose event would be administered indefinitely, so we
                # stop with final_time or 1.
                n_doses = 1

                if np.isfinite(final_time):
                    n_doses = int(abs(final_time) // period)

            # Construct dose times
            dose_times = [start_time + n * period for n in range(n_doses)]

            # Make sure that even for finite periodic dose events the final
            # time is not exceeded
            dose_times = np.array(dose_times)
            mask = dose_times <= final_time
            dose_times = dose_times[mask]

            # Add dose administrations to dataframe
            regimen_df = pd.concat([regimen_df, pd.DataFrame({
                'Time': dose_times,
                'Duration': dose_duration,
                'Dose': dose_amount})])

        # If no dose event before final_time exist, return None
        if regimen_df.empty:
            return None

        return regimen_df

    def get_n_outputs(self):
        """
        Returns the number of outputs.
        """
        return self._mechanistic_model.n_outputs()

    def get_output_names(self):
        """
        Returns the output names.
        """
        return self._mechanistic_model.outputs()

    def get_parameter_names(self):
        """
        Returns the parameter names of the predictive model.
        """
        return copy.copy(self._parameter_names)

    def get_submodels(self):
        """
        Returns the submodels of the predictive model in form of a dictionary.
        """
        # Get original submodels
        mechanistic_model = self._mechanistic_model
        if isinstance(mechanistic_model, chi.ReducedMechanisticModel):
            mechanistic_model = mechanistic_model.mechanistic_model()

        error_models = []
        for error_model in self._error_models:
            # Get original error model
            if isinstance(error_model, chi.ReducedErrorModel):
                error_model = error_model.get_error_model()

            error_models.append(error_model)

        submodels = dict({
            'Mechanistic model': mechanistic_model,
            'Error models': error_models})

        return submodels

    def n_parameters(self):
        """
        Returns the number of parameters of the predictive model.
        """
        return self._n_parameters

    def sample(
            self, parameters, times, n_samples=None, seed=None,
            return_df=True, include_regimen=False, *args, **kwargs):
        """
        Samples "measurements" of the biomarkers from the predictive model and
        returns them in form of a :class:`pandas.DataFrame` or a
        :class:`numpy.ndarray`.

        The mechanistic model is solved for the provided parameters and times,
        and samples around this solution are drawn from the error models for
        each time point.

        The number of samples for each time point can be specified with
        ``n_samples``.

        Parameters
        ----------
        parameters
            An array-like object with the parameter values of the predictive
            model.
        times
            An array-like object with times at which the virtual "measurements"
            are performed.
        n_samples
            The number of virtual "measurements" that are performed at each
            time point. If ``None`` the biomarkers are measured only once
            at each time point.
        seed
            A seed for the pseudo-random number generator or a
            :class:`numpy.random.Generator`.
        return_df
            A boolean flag which determines whether the output is returned as a
            :class:`pandas.DataFrame` or a :class:`numpy.ndarray`. If ``False``
            the samples are returned as a numpy array of shape
            ``(n_outputs, n_times, n_samples)``.
        include_regimen
            A boolean flag which determines whether the dosing regimen
            information is included in the output. If the samples are returned
            as a :class:`numpy.ndarray`, the dosing information is not
            included.
        """
        parameters = np.asarray(parameters)
        if len(parameters) != self._n_parameters:
            raise ValueError(
                'The length of parameters does not match n_parameters.')

        # Sort parameters into mechanistic model params and error params
        n_parameters = self._mechanistic_model.n_parameters()
        mechanistic_params = parameters[:n_parameters]
        error_params = parameters[n_parameters:]

        # Solve mechanistic model
        times = np.sort(times)
        outputs = self._mechanistic_model.simulate(mechanistic_params, times)
        # Create numpy container for samples
        n_outputs = len(outputs)
        n_times = len(times)
        n_samples = n_samples if n_samples is not None else 1
        container = np.empty(shape=(n_outputs, n_times, n_samples))

        # Sample error around mechanistic model outputs
        start_index = 0
        for output_id, error_model in enumerate(self._error_models):
            end_index = start_index + error_model.n_parameters()

            # Sample
            container[output_id, ...] = error_model.sample(
                parameters=error_params[start_index:end_index],
                model_output=outputs[output_id],
                n_samples=n_samples,
                seed=seed)

            # Update start index
            start_index = end_index

        if return_df is False:
            # Return samples in numpy array format
            return container

        # Structure samples in a pandas.DataFrame
        output_names = self._mechanistic_model.outputs()
        sample_ids = np.arange(start=1, stop=n_samples+1)
        samples = pd.DataFrame(
            columns=['ID', 'Time', 'Observable', 'Value'])

        # Fill in all samples at a specific time point at once
        for output_id, name in enumerate(output_names):
            for time_id, time in enumerate(times):
                samples = pd.concat([samples, pd.DataFrame({
                    'ID': sample_ids,
                    'Time': time,
                    'Observable': name,
                    'Value': container[output_id, time_id, :]})])

        # Add dosing regimen information, if set
        final_time = np.max(times)
        regimen = self.get_dosing_regimen(final_time)
        if (regimen is not None) and (include_regimen is True):
            # Add dosing regimen for each sample
            for _id in sample_ids:
                regimen['ID'] = _id
                samples = pd.concat([samples, regimen])

        return samples

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen with which the compound is administered.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the basic time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        .. note::
            This method requires a :class:`MechanisticModel` that supports
            compound administration.

        Parameters
        ----------
        dose
            The amount of the compound that is injected at each administration.
        start
            Start time of the treatment.
        duration
            Duration of dose administration. For a bolus injection, a dose
            duration of 1% of the time unit should suffice. By default the
            duration is set to 0.01 (bolus).
        period
            Periodicity at which doses are administered. If ``None`` the dose
            is administered only once.
        num
            Number of administered doses. If ``None`` and the periodicity of
            the administration is not ``None``, doses are administered
            indefinitely.
        """
        try:
            self._mechanistic_model.set_dosing_regimen(
                dose, start, duration, period, num)
        except AttributeError:
            # This error means that the mechanistic model is a
            # PharmacodynamicModel and therefore no dosing regimen can be set.
            raise AttributeError(
                'The mechanistic model does not support to set dosing '
                'regimens. This may be because the underlying '
                'chi.MechanisticModel is a '
                'chi.PharmacodynamicModel.')


class PopulationPredictiveModel(PredictiveModel):
    r"""
    Implements a model of a data-generating process.

    The model is defined by an instance of a :class:`PredictiveModel` and an
    instance of a :class:`PopulationModel`. The predictive model
    :math:`p(y | \psi, t)` defines the data-generating process for an
    individual in the population with parameters :math:`\psi`. The population
    model :math:`p(\psi | \theta)` defines how the parameters vary across
    individuals in the population.

    As a result, the data-generating process is defined as

    .. math::
        p(y | \theta, t) =
            \int \mathrm{d}\psi \, p(y | \psi, t)\, p(\psi |\theta).

    Extends :class:`PredictiveModel`.

    Parameters
    ----------
    predictive_model
        An instance of a :class:`PredictiveModel`.
    population_model
        An instance of a :class:`PopulationModel`.
    """

    def __init__(self, predictive_model, population_model):
        # Check inputs
        if not isinstance(predictive_model, chi.PredictiveModel):
            raise TypeError(
                'The predictive model has to be an instance of '
                'chi.PredictiveModel.')
        if not isinstance(population_model, chi.PopulationModel):
            raise TypeError(
                'The population model has to be an instance of '
                'chi.PopulationModel.')

        # Get number and names of non-population predictive model
        n_parameters = predictive_model.n_parameters()

        # Check that there is one population model for each model parameter
        if population_model.n_dim() != n_parameters:
            raise ValueError(
                'The dimension of the population model has to be the same as '
                'the number of predictive model parameters.')

        # Remember predictive model and population model
        self._predictive_model = predictive_model
        self._population_model = population_model

    def fix_parameters(self, name_value_dict):
        """
        Fixes the value of model parameters, and effectively removes them as a
        parameter from the model. Fixing the value of a parameter at ``None``,
        sets the parameter free again.

        Parameters
        ----------
        name_value_dict
            A dictionary with model parameter names as keys, and parameter
            values as values.
        """
        # Check type of dictionanry
        try:
            name_value_dict = dict(name_value_dict)
        except (TypeError, ValueError):
            raise ValueError(
                'The name-value dictionary has to be convertable to a python '
                'dictionary.')

        # Convert models to reduced models
        pop_model = self._population_model
        if not isinstance(pop_model, chi.ReducedPopulationModel):
            pop_model = chi.ReducedPopulationModel(pop_model)

        # Fix model parameters
        pop_model.fix_parameters(name_value_dict)

        if pop_model.n_fixed_parameters() == 0:
            pop_model = pop_model.get_population_model()

        # Safe reduced models
        self._population_model = pop_model

    def get_dosing_regimen(self, final_time=None):
        """
        Returns the dosing regimen of the compound in form of a
        :class:`pandas.DataFrame`.

        The dataframe has a time, a duration, and a dose column, which indicate
        the time point and duration of the dose administration in the time
        units of the mechanistic model, :meth:`MechanisticModel.time_unit`. The
        dose column specifies the amount of the compound that is being
        administered in units of the drug amount variable of the mechanistic
        model.

        If an indefinitely administered dosing regimen is set, i.e. a
        finite duration and undefined number of doses, see
        :meth:`set_dosing_regimen`, only the first administration of the
        dose will appear in the dataframe. Alternatively, a final dose time
        ``final_time`` can be provided, up to which the dose events are
        registered.

        If no dosing regimen has been set, ``None`` is returned.

        Parameters
        ----------
        final_time
            Time up to which dose events are registered in the dataframe. If
            ``None``, all dose events are registered, except for indefinite
            dosing regimens. Here, only the first dose event is registered.
        """
        return self._predictive_model.get_dosing_regimen(final_time)

    def get_n_outputs(self):
        """
        Returns the number of outputs.
        """
        return self._predictive_model.get_n_outputs()

    def get_output_names(self):
        """
        Returns the output names.
        """
        return self._predictive_model.get_output_names()

    def get_parameter_names(self):
        """
        Returns the parameter names of the predictive model.
        """
        return self._population_model.get_parameter_names()

    def n_parameters(self):
        """
        Returns the number of parameters of the predictive model.
        """
        return self._population_model.n_parameters()

    def sample(
            self, parameters, times, n_samples=None, seed=None, return_df=True,
            include_regimen=False, covariates=None):
        """
        Samples measurements of the observables from virtual patients.

        Virtual patients are sampled from the population model and measured by
        sampling from the individual-level predictive model. Each virtual
        patient is measured at each of the provided time points.

        The number of virtual patients that is being measured can be specified
        with ``n_samples``.

        If the data-generating process does not depend on covariates, the
        ``covariates`` input is ignored.

        :param parameters: Population model parameters.
        :type parameters: np.ndarry of shape ``(n_parameters,)``
        :param times: Measurement time points.
        :type times: np.ndarray of shape ``(n_times,)``
        :param n_samples: Number of virtual patients.
        :type n_samples: int, optional
        :param seed: Seed for the pseudo-random number generator.
        :type seed: int or :class:`numpy.random.Generator`, optional
        :param return_df: A boolean flag which determines whether the output is
            returned as a :class:`pandas.DataFrame` or a
            :class:`numpy.ndarray`.
        :type return_df: bool, optional
        :param include_regimen: A boolean flag which determines whether the
            dosing regimen information is included in the output. Only possible
            when ``return_df=True``.
        :type include_regimen: bool, optional
        :param covariates: Covariate values, specifying the sampled
            subpopulation.
        :type covariates: List, np.ndarray of shape ``(n_cov,)`` or
            ``(n_samples, n_cov)``, optional
        :rtype: :class:`pandas.DataFrame` or np.ndarray of shape
            ``(n_outputs, n_times, n_samples)``
        """
        # Check inputs
        if not n_samples:
            n_samples = 1
        n_samples = int(n_samples)

        parameters = np.asarray(parameters)
        if len(parameters) != self.n_parameters():
            raise ValueError(
                'The length of parameters does not match n_parameters.')
        if (self._population_model.n_covariates() > 0):
            covariates = np.asarray(covariates)
            if covariates.ndim == 1:
                covariates = covariates[np.newaxis, :]
            n_s, n_c = covariates.shape
            if n_c != self._population_model.n_covariates():
                raise ValueError(
                    'Provided covariates do not match the number of '
                    'covariates.')
            if (n_s > 1) and (n_s != n_samples):
                raise ValueError(
                    'Provided covariates cannot be broadcasted to number of '
                    'samples.')

        if seed is not None:
            seed = np.random.default_rng(seed)

        # Sample individuals from population model
        patients = self._population_model.sample(
            parameters=parameters, n_samples=n_samples, seed=seed,
            covariates=covariates)
        patients = self._population_model.compute_individual_parameters(
            parameters=parameters, eta=patients, covariates=covariates)

        # Create numpy container for samples (measurements of virtual patients)
        n_outputs = self._predictive_model.get_n_outputs()
        n_times = len(times)
        measurements = np.empty(shape=(n_outputs, n_times, n_samples))

        # Sample measurements for each patient
        times = np.sort(times)
        for patient_id, patient in enumerate(patients):
            measurements[..., patient_id] = self._predictive_model.sample(
                parameters=patient, times=times, seed=seed, return_df=False
                )[..., 0]

        if return_df is False:
            # Return samples in numpy array format
            return measurements

        # Structure samples in a pandas.DataFrame
        # (Exploit how .flatten() arranges measurements)
        names = self._predictive_model.get_output_names()
        output_names = []
        for name in names:
            output_names += [name] * (n_times * n_samples)
        times = np.broadcast_to(
            times[np.newaxis, :, np.newaxis],
            shape=(n_outputs, n_times, n_samples)).flatten()
        sample_ids = np.arange(start=1, stop=n_samples+1)
        sample_ids = np.broadcast_to(
            sample_ids[np.newaxis, np.newaxis, :],
            shape=(n_outputs, n_times, n_samples)).flatten()
        measurements = pd.DataFrame({
            'ID': sample_ids,
            'Time': times,
            'Observable': output_names,
            'Value': measurements.flatten()})

        # Append covariates, if used
        if covariates is not None:
            names = self._population_model.get_covariate_names()
            sample_ids = np.arange(start=1, stop=n_samples+1)
            for idc, covariate in enumerate(names):
                measurements = pd.concat([measurements, pd.DataFrame({
                    'ID': sample_ids,
                    'Time': np.nan,
                    'Observable': covariate,
                    'Value': covariates[..., idc]})])

        # Add dosing regimen information, if set
        if include_regimen:
            final_time = np.max(times)
            regimen = self.get_dosing_regimen(final_time)
            if regimen is not None:
                # Add dosing regimen for each sample
                sample_ids = np.arange(start=1, stop=n_samples+1)
                for _id in sample_ids:
                    regimen['ID'] = _id
                    measurements = pd.concat([measurements, regimen])

        return measurements

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen with which the compound is administered.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the basic time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        .. note::
            This method requires a :class:`MechanisticModel` that supports
            compound administration.

        Parameters
        ----------
        dose
            The amount of the compound that is injected at each administration.
        start
            Start time of the treatment.
        duration
            Duration of dose administration. For a bolus injection, a dose
            duration of 1% of the time unit should suffice. By default the
            duration is set to 0.01 (bolus).
        period
            Periodicity at which doses are administered. If ``None`` the dose
            is administered only once.
        num
            Number of administered doses. If ``None`` and the periodicity of
            the administration is not ``None``, doses are administered
            indefinitely.
        """
        self._predictive_model.set_dosing_regimen(
            dose, start, duration, period, num)


class PriorPredictiveModel(AveragedPredictiveModel):
    """
    Implements a model that predicts the change of observable biomarkers over
    time based on the provided distribution of model parameters prior to the
    inference.

    A prior predictive model may be used to check whether the assumptions about
    the parameter distribution ``log_prior`` lead to a predictive distirbution
    that encapsulates the expected measurement values of preclinical and
    clinical biomarkers.

    A PriorPredictiveModel is instantiated with an instance of a
    :class:`PredictiveModel` and a :class:`pints.LogPrior` of the same
    parametric dimension as the predictive model. Future biomarker
    "measurements" can then be predicted by first sampling parameter values
    from the log-prior distribution, and then generating "virtual" measurements
    from the predictive model with those parameters.

    Parameters
    ----------
    predictive_model
        An instance of a :class:`PredictiveModel`.
    log_prior
        An instance of a :class:`pints.LogPrior` of the same dimensionality as
        the number of predictive model parameters.
    """

    def __init__(self, predictive_model, log_prior):
        super(PriorPredictiveModel, self).__init__(predictive_model)

        if not isinstance(log_prior, pints.LogPrior):
            raise ValueError(
                'The provided log-prior has to be an instance of a '
                'pints.LogPrior.')

        if predictive_model.n_parameters() != log_prior.n_parameters():
            raise ValueError(
                'The dimension of the log-prior has to be the same as the '
                'number of parameters of the predictive model.')

        self._log_prior = log_prior

    def sample(
            self, times, n_samples=None, seed=None, include_regimen=False,
            covariates=None):
        """
        Samples "measurements" of the biomarkers from the prior predictive
        model and returns them in form of a :class:`pandas.DataFrame`.

        For each of the ``n_samples`` a parameter set is drawn from the
        log-prior. These paramaters are then used to sample from the predictive
        model.

        :param times: Times for the virtual "measurements".
        :type times: list, numpy.ndarray of shape (n,)
        :param n_samples: The number of virtual "measurements" that are
            performed at each time point. If ``None`` the biomarkers are
            measured only once at each time point.
        :type n_samples: int, optional
        :param seed: Seed for the pseudo-random number generator.
        :type seed: int or :class:`numpy.random.Generator`, optional
        :param include_regimen: A boolean flag which determines whether the
            dosing regimen information is included in the output. Only possible
            when ``return_df=True``.
        :type include_regimen: bool, optional
        :param covariates: Covariate values, specifying the sampled
            subpopulation.
        :type covariates: List, np.ndarray of shape ``(n_cov,)`` or
            ``(n_samples, n_cov)``, optional
        :rtype: :class:`pandas.DataFrame`
        """
        # Make sure n_samples is an integer
        if n_samples is None:
            n_samples = 1
        n_samples = int(n_samples)

        # Set seed for prior samples
        # (does not affect predictive model)
        if seed is not None:
            # TODO: pints.Priors are not meant to be seeded, so fails when
            # anything else but np.random is used.
            np.random.seed(seed)

            # Set predictive model base seed
            base_seed = seed

        # Sort times
        times = np.sort(times)

        # Create container for samples
        container = pd.DataFrame(
            columns=['ID', 'Time', 'Observable', 'Value'])

        # Get model outputs
        outputs = self._predictive_model.get_output_names()

        # Draw samples
        sample_ids = np.arange(start=1, stop=n_samples+1)
        for sample_id in sample_ids:
            # Sample parameter
            parameters = self._log_prior.sample().flatten()

            if seed is not None:
                # Set seed for predictive model to base_seed + sample_id
                # (Needs to change every iteration)
                seed = base_seed + sample_id

            # Sample from predictive model
            sample = self._predictive_model.sample(
                parameters, times, n_samples, seed, return_df=False,
                covariates=covariates)

            # Append samples to dataframe
            for output_id, name in enumerate(outputs):
                container = pd.concat([container, pd.DataFrame({
                    'ID': sample_id,
                    'Time': times,
                    'Observable': name,
                    'Value': sample[output_id, :, 0]})])

        # Add dosing regimen, if set
        final_time = np.max(times)
        regimen = self.get_dosing_regimen(final_time)
        if (regimen is not None) and (include_regimen is True):
            # Append dosing regimen only once for all samples
            container = pd.concat([container, regimen])

        return container


class PAMPredictiveModel(AveragedPredictiveModel):
    r"""
    A model that is defined by the probabilistic average of
    posterior predictive models.

    Probabilistic averging of models is the weighted average of the predictive
    distributions of individual models

    .. math::
        p(x | x^{\mathrm{obs}}) = \sum _m w_m\, p_m(x | x^{\mathrm{obs}}),

    where the sum runs over the individual models and :math:`w_m` is the weight
    of model :math:`m`.

    .. warning::
        Does currently not support CovariatePopulationModels.

    :param predictive_models: A list of predictive models.
    :type predictive_models: List[PosteriorPredictiveModel] of length
        `n_models`.
    :param weights: The weights of candidate models. Weights are normalised
        automatically.
    :type weights: List np.ndarray of length `n_models`.
    """
    def __init__(self, predictive_models, weights):
        # Check inputs
        for predictive_model in predictive_models:
            if not isinstance(predictive_model, PosteriorPredictiveModel):
                raise TypeError(
                    'The predictive models must be instances of '
                    'chi.PosteriorPredictiveModel.')

        predictive_model = predictive_models[0].get_predictive_model()
        super(PAMPredictiveModel, self).__init__(predictive_model)

        n_outputs = self._predictive_model.get_n_outputs()
        for predictive_model in predictive_models:
            if n_outputs != predictive_model.get_n_outputs():
                raise ValueError(
                    'All predictive models must have the same number of '
                    'outputs.')

        output_names = self._predictive_model.get_output_names()
        for predictive_model in predictive_models:
            if output_names != predictive_model.get_output_names():
                raise Warning(
                    'The predictive models appear to have different output '
                    'names. Stacking of the predictive distributions might '
                    'therefore not meaningful.')

        if len(predictive_models) != len(weights):
            raise ValueError(
                'The model weights must be of the same length as the number '
                'of predictive models.')

        weights = np.array(weights, dtype=float)

        # Remember models and normalised weights
        self._predictive_models = predictive_models
        self._weights = weights / np.sum(weights)

    def get_predictive_model(self):
        """
        Returns a list of the
        :class:`chi.PosteriorPredictiveModel` instances.
        """
        return self._predictive_models

    def get_weights(self):
        """
        Returns the weights of the individual predictive models.
        """
        return copy.copy(self._weights)

    def sample(
            self, times, n_samples=None, individual=None, seed=None,
            include_regimen=False):
        """
        Samples "measurements" of the biomarkers from the posterior predictive
        model and returns them in form of a :class:`pandas.DataFrame`.

        For each of the ``n_samples`` a parameter set is drawn from the
        approximate posterior distribution. These paramaters are then used to
        sample from the predictive model.

        :param times: Times for the virtual "measurements".
        :type times: list, numpy.ndarray of shape (n,)
        :param n_samples: The number of virtual "measurements" that are
            performed at each time point. If ``None`` the biomarkers are
            measured only once at each time point.
        :type n_samples: int, optional
        :param individual: The ID of the modelled individual. If
            ``None``, either the first ID or the population is simulated.
        :type individual: str, optional
        :param seed: A seed for the pseudo-random number generator.
        :type seed: int
        :param include_regimen: A boolean flag which determines whether the
            information about the dosing regimen is included.
        :type include_regimen: bool, optional
        """
        # Make sure n_samples is an integer
        if n_samples is None:
            n_samples = 1
        n_samples = int(n_samples)

        # Instantiate random number generator
        seed = np.random.default_rng(seed=seed)

        # Sample number of samples from each predictive model
        n_models = len(self._predictive_models)
        model_indices = np.arange(n_models)
        model_draws = np.random.choice(
            model_indices, p=self._weights, size=n_samples)
        samples_per_model = np.zeros(n_models, dtype=int)
        for model_id, model in enumerate(model_indices):
            samples_per_model[model_id] = np.sum(
                model_draws == model)

        # Sample from predictive models
        samples = []
        for model_id, n_samples in enumerate(samples_per_model):
            # Skip model if no samples are drawn from it
            if n_samples == 0:
                continue

            # Sample
            model = self._predictive_models[model_id]
            s = model.sample(times, n_samples, individual, seed=seed)

            # Shift IDs by number of previous draws
            s['ID'] += int(np.sum(samples_per_model[:model_id]))

            # Append samples to list
            samples.append(s)

        # Concatenate all samples to one dataframe
        samples = pd.concat(samples)

        # Add dosing regimen, if set
        final_time = np.max(times)
        regimen = self.get_dosing_regimen(final_time)
        if (regimen is not None) and (include_regimen is True):
            # Append dosing regimen only once for all samples
            samples = pd.concat([samples, regimen])

        return samples

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen with which the compound is administered.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the basic time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        .. note::
            This method requires a :class:`MechanisticModel` that supports
            dose administration.

        Parameters
        ----------
        dose
            The amount of the compound that is injected at each administration.
        start
            Start time of the treatment.
        duration
            Duration of dose administration. For a bolus injection, a dose
            duration of 1% of the time unit should suffice. By default the
            duration is set to 0.01 (bolus).
        period
            Periodicity at which doses are administered. If ``None`` the dose
            is administered only once.
        num
            Number of administered doses. If ``None`` and the periodicity of
            the administration is not ``None``, doses are administered
            indefinitely.
        """
        for predictive_model in self._predictive_models:
            predictive_model.set_dosing_regimen(
                dose, start, duration, period, num)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import copy

import myokit
import myokit.formats.sbml as sbml
import numpy as np


class MechanisticModel(object):
    r"""
    A base class for time series models of the form

    .. math::
        \bar{y} = g(\bar{y}, \psi, t),

    where :math:`\bar{y}` are the model outputs, :math:`\psi` are the model
    parameters and :math:`t` is the time. :math:`g` can be any (deterministic)
    function.
    """
    def __init__(self):
        super(MechanisticModel, self).__init__()

    def copy(self):
        """
        Returns a deep copy of the mechanistic model.

        .. note:
            The default implementation may have to be adapted based on the
            implementation of :meth:`simulate`.
        """
        return copy.deepcopy(self)

    def enable_sensitivities(self, enabled, parameter_names=None):
        r"""
        Enables the computation of the model output sensitivities to the model
        parameters if set to ``True``.

        The sensitivities of the model outputs are defined as the partial
        derviatives of the ouputs :math:`\bar{y}` with respect to the model
        parameters :math:`\psi`

        .. math:
            \frac{\del \bar{y}}{\del \psi}.

        :param enabled: A boolean flag which enables (``True``) / disables
            (``False``) the computation of sensitivities.
        :type enabled: bool
        """
        raise NotImplementedError

    def has_sensitivities(self):
        """
        Returns a boolean indicating whether sensitivities have been enabled.
        """
        raise NotImplementedError

    def n_outputs(self):
        """
        Returns the number of output dimensions.

        By default this is the number of states.
        """
        raise NotImplementedError

    def n_parameters(self):
        """
        Returns the number of parameters in the model.

        Parameters of the model are initial state values and structural
        parameter values.
        """
        raise NotImplementedError

    def outputs(self):
        """
        Returns the output names of the model.
        """
        raise NotImplementedError

    def parameters(self):
        """
        Returns the parameter names of the model.
        """
        raise NotImplementedError

    def simulate(self, parameters, times):
        """
        Returns the numerical solution of the model outputs (and optionally
        the sensitivites) for the specified parameters and times.

        The model outputs are returned as a 2 dimensional NumPy array of shape
        ``(n_outputs, n_times)``. If sensitivities are enabled, a tuple is
        returned with the NumPy array of the model outputs and a NumPy array of
        the sensitivities of shape ``(n_times, n_outputs, n_parameters)``.

        :param parameters: An array-like object with values for the model
            parameters.
        :type parameters: list, numpy.ndarray
        :param times: An array-like object with time points at which the output
            values are returned.
        :type times: list, numpy.ndarray

        :rtype: np.ndarray of shape (n_outputs, n_times) or
            (n_times, n_outputs, n_parameters)
        """
        raise NotImplementedError

    def supports_dosing(self):
        """
        Returns a boolean whether dose administration with
        :meth:`PKPDModel.set_dosing_regimen` is supported by the model.
        """
        return False


class SBMLModel(MechanisticModel):
    """
    Instantiates a mechanistic model from a SBML specification.

    Extends :class:`MechanisticModel`.

    :param sbml_file: A path to the SBML model file that specifies the model.
    :type sbml_file: str
    """

    def __init__(self, sbml_file):
        super(SBMLModel, self).__init__()

        # Import model
        self._model = sbml.SBMLImporter().model(sbml_file)

        # Set default number and names of states, parameters and outputs.
        self._set_number_and_names()

        # Get time unit
        self._time_unit = self._get_time_unit()

        # Create simulator without sensitivities
        # (intentionally public property)
        self._simulator = myokit.Simulation(self._model)
        self._has_sensitivities = False

    def _get_time_unit(self):
        """
        Gets the model's time unit.
        """
        # Get bound variables
        bound_variables = [var for var in self._model.variables(bound=True)]

        # Get the variable that is bound to time
        # (only one can exist in myokit.Model)
        for var in bound_variables:
            if var._binding == 'time':
                return var.unit()

    def _set_const(self, parameters):
        """
        Sets values of constant model parameters.
        """
        for id_var, var in enumerate(self._const_names):
            self._simulator.set_constant(var, float(parameters[id_var]))

    def _set_state(self, parameters):
        """
        Sets initial values of states.
        """
        parameters = np.array(parameters)
        parameters = parameters[self._original_order]
        self._simulator.set_state(parameters)

    def _set_number_and_names(self):
        """
        Sets the number of states, parameters and outputs, as well as their
        names. If the model is ``None`` the self._model is taken.
        """
        # Get the number of states and parameters
        self._n_states = self._model.count_states()
        n_const = self._model.count_variables(const=True)

        # Get constant variable names and state names
        names = [var.qname() for var in self._model.states()]
        self._state_names = sorted(names)

        const_names = []
        for var in self._model.variables(const=True):
            # Sometimes constants are derived from parameters
            if not var.is_literal():
                n_const -= 1
                continue
            const_names.append(var.qname())
        self._const_names = sorted(const_names)
        self._n_parameters = self._n_states + n_const

        # Remember original order of state names for simulation
        order_after_sort = np.argsort(names)
        self._original_order = np.argsort(order_after_sort)

        # Set default parameter names
        self._parameter_names = self._state_names + self._const_names

        # Temporarily set outputs to default outputs
        self._output_names = self._state_names
        self._n_outputs = self._n_states

        # Create references of displayed parameter and output names to
        # orginal myokit names (defaults to identity map)
        # (Key: myokit name, value: displayed name)
        self._parameter_name_map = dict(
            zip(self._parameter_names, self._parameter_names))
        self._output_name_map = dict(
            zip(self._output_names, self._output_names))

    def copy(self):
        """
        Returns a deep copy of the mechanistic model.

        .. note::
            Copying the model resets the sensitivity settings.
        """
        # Copy model manually and get protocol
        m = self._model.clone()
        s = self._simulator
        myokit_model = m.clone()
        self._model = None
        self._simulator = None

        # Copy the mechanistic model
        model = copy.deepcopy(self)

        # Replace myokit model by safe copy and create simulator
        self._model = m
        self._simulator = s
        model._model = myokit_model
        model._simulator = myokit.Simulation(myokit_model)

        return model

    def enable_sensitivities(self, enabled, parameter_names=None):
        """
        Enables the computation of the model output sensitivities to the model
        parameters if set to ``True``.

        The sensitivities are computed using the forward sensitivities method,
        where an ODE for each sensitivity is derived. The sensitivities are
        returned together with the solution to the orginal system of ODEs when
        simulating the mechanistic model :meth:`simulate`.

        The optional parameter names argument can be used to set which
        sensitivities are computed. By default the sensitivities to all
        parameters are computed.

        :param enabled: A boolean flag which enables (``True``) / disables
            (``False``) the computation of sensitivities.
        :type enabled: bool
        :param parameter_names: A list of parameter names of the model. If
            ``None`` sensitivities for all parameters are computed.
        :type parameter_names: list[str], optional
        """
        enabled = bool(enabled)

        if not enabled:
            if self._has_sensitivities:
                # Disable sensitivities
                sim = myokit.Simulation(self._model)
                self._simulator = sim
                self._has_sensitivities = False

                return None

            # Sensitivities are already disabled
            return None

        # Get parameters whose output sensitivities are computed
        parameters = []
        for param_id, param in enumerate(self._parameter_names):
            if param_id < self._n_states:
                # Convert initial value parameters to the correct syntax
                parameters.append('init(' + param + ')')
                continue

            # Other parameters can be appended without modification
            parameters.append(param)

        if parameter_names is not None:
            # Get myokit names for input parameter names
            container = []
            for index, public_name in enumerate(
                    self._parameter_name_map.values()):
                if public_name in parameter_names:
                    container.append(parameters[index])

            parameters = container

        if not parameters:
            raise ValueError(
                'None of the parameters could be identified. The valid '
                'parameter names are <' + str(self._parameter_names) + '>.')

        # Create simulator
        sensitivities = (self._output_names, parameters)
        sim = myokit.Simulation(
            self._model, protocol=None, sensitivities=sensitivities)

        # Update simulator and sensitivity state
        self._simulator = sim
        self._has_sensitivities = True

    def has_sensitivities(self):
        """
        Returns a boolean indicating whether sensitivities have been enabled.
        """
        return self._has_sensitivities

    def n_outputs(self):
        """
        Returns the number of output dimensions.

        By default this is the number of states.
        """
        return self._n_outputs

    def n_parameters(self):
        """
        Returns the number of parameters in the model.

        Parameters of the model are initial state values and structural
        parameter values.
        """
        return self._n_parameters

    def outputs(self):
        """
        Returns the output names of the model.
        """
        # Get user specified output names
        output_names = [
            self._output_name_map[name] for name in self._output_names]
        return output_names

    def parameters(self):
        """
        Returns the parameter names of the model.
        """
        # Get user specified parameter names
        parameter_names = [
            self._parameter_name_map[name] for name in self._parameter_names]

        return parameter_names

    def set_outputs(self, outputs):
        """
        Sets outputs of the model.

        The outputs can be set to any quantifiable variable name of the
        :class:`myokit.Model`, e.g. `compartment.variable`.

        .. note::
            Setting outputs resets the sensitivity settings (by default
            sensitivities are disabled.)

        :param outputs:
            A list of output names.
        :type outputs: list[str]
        """
        outputs = list(outputs)

        # Translate public names to myokit names, if set previously
        for myokit_name, public_name in self._output_name_map.items():
            if public_name in outputs:
                # Replace public name by myokit name
                index = outputs.index(public_name)
                outputs[index] = myokit_name

        # Check that outputs are valid
        for output in outputs:
            try:
                var = self._simulator._model.get(output)
                if not (var.is_state() or var.is_intermediary()):
                    raise ValueError(
                        'Outputs have to be state or intermediary variables.')
            except KeyError:
                raise KeyError(
                    'The variable <' + str(output) + '> does not exist in the '
                    'model.')

        # Remember outputs
        self._output_names = outputs
        self._n_outputs = len(outputs)

        # Create an updated output name map
        output_name_map = {}
        for myokit_name in self._output_names:
            try:
                output_name_map[myokit_name] = self._output_name_map[
                    myokit_name]
            except KeyError:
                # The output did not exist before, so create an identity map
                output_name_map[myokit_name] = myokit_name
        self._output_name_map = output_name_map

        # Disable sensitivities
        self.enable_sensitivities(False)

    def set_output_names(self, names):
        """
        Assigns names to the model outputs. By default the
        :class:`myokit.Model` names are assigned to the outputs.

        :param names: A dictionary that maps the current output names to new
            names.
        :type names: dict[str, str]
        """
        if not isinstance(names, dict):
            raise TypeError(
                'Names has to be a dictionary with the current output names'
                'as keys and the new output names as values.')

        # Check that new output names are unique
        new_names = list(names.values())
        n_unique_new_names = len(set(names.values()))
        if len(new_names) != n_unique_new_names:
            raise ValueError(
                'The new output names have to be unique.')

        # Check that new output names do not exist already
        for new_name in new_names:
            if new_name in list(self._output_name_map.values()):
                raise ValueError(
                    'The output names cannot coincide with existing '
                    'output names. One output is already called '
                    '<' + str(new_name) + '>.')

        # Replace currently displayed names by new names
        for myokit_name in self._output_names:
            old_name = self._output_name_map[myokit_name]
            try:
                new_name = names[old_name]
                self._output_name_map[myokit_name] = str(new_name)
            except KeyError:
                # KeyError indicates that the current output is not being
                # renamed.
                pass

    def set_parameter_names(self, names):
        """
        Assigns names to the parameters. By default the :class:`myokit.Model`
        names are assigned to the parameters.

        :param names: A dictionary that maps the current parameter names to new
            names.
        :type names: dict[str, str]
        """
        if not isinstance(names, dict):
            raise TypeError(
                'Names has to be a dictionary with the current parameter names'
                'as keys and the new parameter names as values.')

        # Check that new parameter names are unique
        new_names = list(names.values())
        n_unique_new_names = len(set(names.values()))
        if len(new_names) != n_unique_new_names:
            raise ValueError(
                'The new parameter names have to be unique.')

        # Check that new parameter names do not exist already
        for new_name in new_names:
            if new_name in list(self._parameter_name_map.values()):
                raise ValueError(
                    'The parameter names cannot coincide with existing '
                    'parameter names. One parameter is already called '
                    '<' + str(new_name) + '>.')

        # Replace currently displayed names by new names
        for myokit_name in self._parameter_names:
            old_name = self._parameter_name_map[myokit_name]
            try:
                new_name = names[old_name]
                self._parameter_name_map[myokit_name] = str(new_name)
            except KeyError:
                # KeyError indicates that the current parameter is not being
                # renamed.
                pass

    def simulate(self, parameters, times):
        """
        Returns the numerical solution of the model outputs (and optionally
        the sensitivites) for the specified parameters and times.

        The model outputs are returned as a 2 dimensional NumPy array of shape
        ``(n_outputs, n_times)``. If sensitivities are enabled, a tuple is
        returned with the NumPy array of the model outputs and a NumPy array of
        the sensitivities of shape ``(n_times, n_outputs, n_parameters)``.

        :param parameters: An array-like object with values for the model
            parameters.
        :type parameters: list, numpy.ndarray
        :param times: An array-like object with time points at which the output
            values are returned.
        :type times: list, numpy.ndarray

        :rtype: np.ndarray of shape (n_outputs, n_times) or a Tuple of two
            np.ndarray, the first of shape (n_outputs, n_times) and the second
            of shape (n_times, n_outputs, n_parameters)
        """
        # Reset simulation
        self._simulator.reset()

        # Set initial conditions
        self._set_state(parameters[:self._n_states])

        # Set constant model parameters
        self._set_const(parameters[self._n_states:])

        # Simulate
        if not self._has_sensitivities:
            output = self._simulator.run(
                times[-1] + 1, log=self._output_names, log_times=times)
            output = np.array([output[name] for name in self._output_names])

            return output

        output, sensitivities = self._simulator.run(
            times[-1] + 1, log=self._output_names, log_times=times)
        output = np.array([output[name] for name in self._output_names])
        sensitivities = np.array(sensitivities)

        return output, sensitivities

    def time_unit(self):
        """
        Returns the model's unit of time.
        """
        return self._time_unit


class PKPDModel(SBMLModel):
    """
    Instantiates a PKPD model from a SBML specification.

    Extends :class:`SBMLModel`.

    :param sbml_file: A path to the SBML model file that specifies the
        PKPD model.
    :type sbml_file: str
    """

    def __init__(self, sbml_file):
        super(PKPDModel, self).__init__(sbml_file)

        # Set default dose administration
        self._administration = None
        self._dosing_regimen = None

        # Safe vanilla model
        self._vanilla_model = self._model.clone()

    def _add_dose_compartment(self, model, drug_amount):
        """
        Adds a dose compartment to the model with a linear absorption rate to
        the connected compartment.
        """
        # Add a dose compartment to the model
        dose_comp = model.add_component_allow_renaming('dose')

        # Create a state variable for the drug amount in the dose compartment
        dose_drug_amount = dose_comp.add_variable('drug_amount')
        dose_drug_amount.set_rhs(0)
        dose_drug_amount.set_unit(drug_amount.unit())
        dose_drug_amount.promote()

        # Create an absorption rate variable
        absorption_rate = dose_comp.add_variable('absorption_rate')
        absorption_rate.set_rhs(1)
        absorption_rate.set_unit(1 / self.time_unit())

        # Add outflow expression to dose compartment
        dose_drug_amount.set_rhs(
            myokit.Multiply(
                myokit.PrefixMinus(myokit.Name(absorption_rate)),
                myokit.Name(dose_drug_amount)
                )
            )

        # Add inflow expression to connected compartment
        rhs = drug_amount.rhs()
        drug_amount.set_rhs(
            myokit.Plus(
                rhs,
                myokit.Multiply(
                    myokit.Name(absorption_rate),
                    myokit.Name(dose_drug_amount)
                )
            )
        )

        # Update number of parameters and states, as well as their names
        # (This overwrites current outputs, so we have to set them again)
        self._model = model
        original_outputs = self._output_names
        self._set_number_and_names()
        self.set_outputs(original_outputs)

        return model, dose_drug_amount

    def _add_dose_rate(self, compartment, drug_amount):
        """
        Adds a dose rate variable to the state variable, which is bound to the
        dosing regimen.
        """
        # Register a dose rate variable to the compartment and bind it to
        # pace, i.e. tell myokit that its value is set by the dosing regimen/
        # myokit.Protocol
        dose_rate = compartment.add_variable_allow_renaming(
            str('dose_rate'))
        dose_rate.set_binding('pace')

        # Set initial value to 0 and unit to unit of drug amount over unit of
        # time
        dose_rate.set_rhs(0)
        if drug_amount.unit():
            dose_rate.set_unit(drug_amount.unit() / self.time_unit())

        # Add the dose rate to the rhs of the drug amount variable
        rhs = drug_amount.rhs()
        drug_amount.set_rhs(
            myokit.Plus(
                rhs,
                myokit.Name(dose_rate)
            )
        )

    def administration(self):
        """
        Returns the mode of administration in form of a dictionary.

        The dictionary has the keys 'compartment' and 'direct'. The former
        provides information about which compartment is dosed, and the latter
        whether the dose is administered directly ot indirectly to the
        compartment.
        """
        return self._administration

    def copy(self):
        """
        Returns a deep copy of the mechanistic model.

        .. note::
            Copying the model resets the sensitivity settings.
        """
        model = super(PKPDModel, self).copy()
        model._simulator.set_protocol(model.dosing_regimen())

        return model

    def dosing_regimen(self):
        """
        Returns the dosing regimen of the compound in form of a
        :class:`myokit.Protocol`. If the protocol has not been set, ``None`` is
        returned.
        """
        return self._dosing_regimen

    def enable_sensitivities(self, enabled, parameter_names=None):
        """
        Enables the computation of the model output sensitivities to the model
        parameters if set to ``True``.

        The sensitivities are computed using the forward sensitivities method,
        where an ODE for each sensitivity is derived. The sensitivities are
        returned together with the solution to the orginal system of ODEs when
        simulating the mechanistic model :meth:`simulate`.

        The optional parameter names argument can be used to set which
        sensitivities are computed. By default the sensitivities to all
        parameters are computed.

        :param enabled: A boolean flag which enables (``True``) / disables
            (``False``) the computation of sensitivities.
        :type enabled: bool
        :param parameter_names: A list of parameter names of the model. If
            ``None`` sensitivities for all parameters are computed.
        :type parameter_names: list[str], optional
        """
        enabled = bool(enabled)

        # Check whether myokit.Simulation needs to be updated
        new_sim = False
        if enabled or ((not enabled) and self._has_sensitivities):
            new_sim = True

        # Set sensitivities
        super(PKPDModel, self).enable_sensitivities(enabled, parameter_names)

        # Update dosing regimen if sensitivity has resulted in new
        # myokit.Simulation instance
        if new_sim:
            self._simulator.set_protocol(self._dosing_regimen)

    def set_administration(
            self, compartment, amount_var='drug_amount', direct=True):
        r"""
        Sets the route of administration of the compound.

        The compound is administered to the selected compartment either
        directly or indirectly. If it is administered directly, a dose rate
        variable is added to the drug amount's rate of change expression

        .. math ::

            \frac{\text{d}A}{\text{d}t} = \text{RHS} + r_d,

        where :math:`A` is the drug amount in the selected compartment, RHS is
        the rate of change of :math:`A` prior to adding the dose rate, and
        :math:`r_d` is the dose rate.

        The dose rate can be set by :meth:`set_dosing_regimen`.

        If the route of administration is indirect, a dosing compartment
        is added to the model, which is connected to the selected compartment.
        The dose rate variable is then added to the rate of change expression
        of the dose amount variable in the dosing compartment. The drug amount
        in the dosing compartment flows at a linear absorption rate into the
        selected compartment

        .. math ::

            \frac{\text{d}A_d}{\text{d}t} = -k_aA_d + r_d \\
            \frac{\text{d}A}{\text{d}t} = \text{RHS} + k_aA_d,

        where :math:`A_d` is the amount of drug in the dose compartment and
        :math:`k_a` is the absorption rate.

        Setting an indirect administration route changes the number of
        parameters of the model, because an initial dose compartment drug
        amount and a absorption rate parameter are added.

        .. note:
            Setting the route of administration will reset the sensitivity
            settings.

        :param compartment: Compartment to which doses are either directly or
            indirectly administered.
        :type compartment: str
        :param amount_var: Drug amount variable in the compartment. By default
            the drug amount variable is assumed to be 'drug_amount'.
        :type amount_var: str, optional
        :param direct: A boolean flag that indicates whether the dose is
            administered directly or indirectly to the compartment.
        :type direct: bool, optional
        """
        # Check inputs
        model = self._vanilla_model.clone()
        if not model.has_component(compartment):
            raise ValueError(
                'The model does not have a compartment named <'
                + str(compartment) + '>.')
        comp = model.get(compartment, class_filter=myokit.Component)

        if not comp.has_variable(amount_var):
            raise ValueError(
                'The drug amount variable <' + str(amount_var) + '> could not '
                'be found in the compartment.')

        drug_amount = comp.get(amount_var)
        if not drug_amount.is_state():
            raise ValueError(
                'The variable <' + str(drug_amount) + '> is not a state '
                'variable, and can therefore not be dosed.')

        # If administration is indirect, add a dosing compartment and update
        # the drug amount variable to the one in the dosing compartment
        if not direct:
            model, drug_amount = self._add_dose_compartment(model, drug_amount)
            comp = model.get(compartment, class_filter=myokit.Component)

        # Add dose rate variable to the right hand side of the drug amount
        self._add_dose_rate(comp, drug_amount)

        # Update model and simulator
        # (otherwise simulator won't know about pace bound variable)
        self._model = model
        self._simulator = myokit.Simulation(model)
        self._has_sensitivities = False

        # Remember type of administration
        self._administration = dict(
            {'compartment': compartment, 'direct': direct})

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen with which the compound is administered.

        The route of administration can be set with :meth:`set_administration`.
        However, the type of administration, e.g. bolus injection or infusion,
        may be controlled with the duration input.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the basic time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        :param dose: The amount of the compound that is injected at each
            administration, or a myokit.Protocol instance that defines the
            dosing regimen.
        :type dose: float or myokit.Protocol
        :param start: Start time of the treatment. By default the
            administration starts at t=0.
        :type start: float, optional
        :param duration: Duration of dose administration. By default the
            duration is set to 0.01 of the time unit (bolus).
        :type duration: float, optional
        :param period: Periodicity at which doses are administered. If ``None``
            the dose is administered only once.
        :type period: float, optional
        :param num: Number of administered doses. If ``None`` and the
            periodicity of the administration is not ``None``, doses are
            administered indefinitely.
        :type num: int, optional
        """
        if self._administration is None:
            raise ValueError(
                'The route of administration of the dose has not been set.')

        if num is None:
            # Myokits default is zero, i.e. infinitely many doses
            num = 0

        if period is None:
            # If period is not provided, we administer a single dose
            # Myokits defaults are 0s for that.
            period = 0
            num = 0

        if isinstance(dose, myokit.Protocol):
            self._simulator.set_protocol(dose)
            self._dosing_regimen = dose
            return None

        # Translate dose to dose rate
        dose_rate = dose / duration

        # Set dosing regimen
        dosing_regimen = myokit.pacing.blocktrain(
            period=period, duration=duration, offset=start, level=dose_rate,
            limit=num)
        self._simulator.set_protocol(dosing_regimen)
        self._dosing_regimen = dosing_regimen

    def supports_dosing(self):
        """
        Returns a boolean whether dose administration with
        :meth:`PKPDModel.set_dosing_regimen` is supported by the model.
        """
        return True


class ReducedMechanisticModel(MechanisticModel):
    """
    A wrapper class for a :class:`MechanisticModel` instance that can be used
    to fix model parameters to fixed values.

    Extends :class:`MechanisticModel`.

    :param mechanistic_model: A mechanistic model.
    :type mechanistic_model: chi.MechanisticModel
    """
    def __init__(self, mechanistic_model):
        super(ReducedMechanisticModel, self).__init__()

        # Check input
        if not isinstance(mechanistic_model, MechanisticModel):
            raise ValueError(
                'The mechanistic model has to be an instance of a '
                'chi.MechanisticModel')

        self._mechanistic_model = mechanistic_model

        # Set defaults
        self._fixed_params_mask = None
        self._fixed_params_values = None
        self._n_parameters = mechanistic_model.n_parameters()
        self._parameter_names = mechanistic_model.parameters()

    def copy(self):
        """
        Returns a deep copy of the reduced model.

        .. note::
            Copying the model resets the sensitivity settings.
        """
        # Get a safe copy of the mechanistic model
        m = self._mechanistic_model
        mechanistic_model = m.copy()
        self._mechanistic_model = None

        # Copy the reduced model
        # (this possibly corrupts the mechanistic model and the
        # simulator)
        model = copy.deepcopy(self)

        # Replace mechanistic model
        self._mechanistic_model = m
        model._mechanistic_model = mechanistic_model

        return model

    def dosing_regimen(self):
        """
        Returns the dosing regimen of the compound in form of a
        :class:`myokit.Protocol`. If the protocol has not been set, ``None`` is
        returned.

        If the model does not support dose administration, ``None`` is
        returned.
        """
        try:
            return self._mechanistic_model.dosing_regimen()
        except AttributeError:
            return None

    def enable_sensitivities(self, enabled):
        """
        Enables the computation of the output sensitivities with respect to
        the free model parameters.
        """
        if not enabled:
            self._mechanistic_model.enable_sensitivities(enabled)
            return None

        # Get free parameters
        free_parameters = np.array(self._parameter_names)
        if self._fixed_params_mask is not None:
            free_parameters = free_parameters[~self._fixed_params_mask]

        # Set sensitivities
        self._mechanistic_model.enable_sensitivities(
            enabled, free_parameters)

    def fix_parameters(self, name_value_dict):
        """
        Fixes the value of model parameters, and effectively removes them as a
        parameter from the model. Fixing the value of a parameter at ``None``,
        sets the parameter free again.

        Parameters
        ----------
        name_value_dict
            A dictionary with model parameter names as keys, and parameter
            values as values.
        """
        # Check type
        try:
            name_value_dict = dict(name_value_dict)
        except (TypeError, ValueError):
            raise ValueError(
                'The name-value dictionary has to be convertable to a python '
                'dictionary.')

        # If no model parameters have been fixed before, instantiate a mask
        # and values
        if self._fixed_params_mask is None:
            self._fixed_params_mask = np.zeros(
                shape=self._n_parameters, dtype=bool)

        if self._fixed_params_values is None:
            self._fixed_params_values = np.empty(shape=self._n_parameters)

        # Update the mask and values
        for index, name in enumerate(self._parameter_names):
            try:
                value = name_value_dict[name]
            except KeyError:
                # KeyError indicates that parameter name is not being fixed
                continue

            # Fix parameter if value is not None, else unfix it
            self._fixed_params_mask[index] = value is not None
            self._fixed_params_values[index] = value

        # If all parameters are free, set mask and values to None again
        if np.all(~self._fixed_params_mask):
            self._fixed_params_mask = None
            self._fixed_params_values = None

        # Remove sensitivities for fixed parameters
        if self.has_sensitivities() is True:
            self.enable_sensitivities(True)

    def has_sensitivities(self):
        """
        Returns a boolean indicating whether sensitivities have been enabled.
        """
        return self._mechanistic_model.has_sensitivities()

    def mechanistic_model(self):
        """
        Returns the original mechanistic model.
        """
        return self._mechanistic_model

    def n_fixed_parameters(self):
        """
        Returns the number of fixed model parameters.
        """
        if self._fixed_params_mask is None:
            return 0

        n_fixed = int(np.sum(self._fixed_params_mask))

        return n_fixed

    def n_outputs(self):
        """
        Returns the number of output dimensions.

        By default this is the number of states.
        """
        return self._mechanistic_model.n_outputs()

    def n_parameters(self):
        """
        Returns the number of parameters in the model.

        Parameters of the model are initial state values and structural
        parameter values.
        """
        # Get number of fixed parameters
        n_fixed = 0
        if self._fixed_params_mask is not None:
            n_fixed = int(np.sum(self._fixed_params_mask))

        # Subtract fixed parameters from total number
        n_parameters = self._n_parameters - n_fixed

        return n_parameters

    def outputs(self):
        """
        Returns the output names of the model.
        """
        return self._mechanistic_model.outputs()

    def parameters(self):
        """
        Returns the parameter names of the model.
        """
        # Remove fixed model parameters
        names = self._parameter_names
        if self._fixed_params_mask is not None:
            names = np.array(names)
            names = names[~self._fixed_params_mask]
            names = list(names)

        return copy.copy(names)

    def set_dosing_regimen(
            self, dose, start=0, duration=0.01, period=None, num=None):
        """
        Sets the dosing regimen with which the compound is administered.

        The route of administration can be set with :meth:`set_administration`.
        However, the type of administration, e.g. bolus injection or infusion,
        may be controlled with the duration input.

        By default the dose is administered as a bolus injection (duration on
        a time scale that is 100 fold smaller than the basic time unit). To
        model an infusion of the dose over a longer time period, the
        ``duration`` can be adjusted to the appropriate time scale.

        By default the dose is administered once. To apply multiple doses
        provide a dose administration period.

        Parameters
        ----------
        dose
            The amount of the compound that is injected at each administration.
        start
            Start time of the treatment.
        duration
            Duration of dose administration. For a bolus injection, a dose
            duration of 1% of the time unit should suffice. By default the
            duration is set to 0.01 (bolus).
        period
            Periodicity at which doses are administered. If ``None`` the dose
            is administered only once.
        num
            Number of administered doses. If ``None`` and the periodicity of
            the administration is not ``None``, doses are administered
            indefinitely.
        """
        try:
            self._mechanistic_model.set_dosing_regimen(
                dose, start, duration, period, num)
        except AttributeError:
            raise AttributeError(
                'The mechanistic model does not support dosing regimens.')

    def set_outputs(self, outputs):
        """
        Sets outputs of the model.

        Parameters
        ----------
        outputs
            A list of quantifiable variable names of the :class:`myokit.Model`,
            e.g. `compartment.variable`.
        """
        try:
            self._mechanistic_model.set_outputs(outputs)
        except AttributeError:
            raise NotImplementedError(
                'The mechanistic model has no implemented set_outputs '
                'method.')

    def set_output_names(self, names):
        """
        Assigns names to the outputs. By default the :class:`myokit.Model`
        names are assigned to the outputs.

        Parameters
        ----------
        names
            A dictionary that maps the current output names to new names.
        """
        try:
            self._mechanistic_model.set_output_names(names)
        except AttributeError:
            raise NotImplementedError(
                'The mechanistic model has no implemented set_output_names '
                'method.')

    def set_parameter_names(self, names):
        """
        Assigns names to the parameters. By default the :class:`myokit.Model`
        names are assigned to the parameters.

        Parameters
        ----------
        names
            A dictionary that maps the current parameter names to new names.
        """
        try:
            self._mechanistic_model.set_parameter_names(names)
            self._parameter_names = self._mechanistic_model.parameters()
        except AttributeError:
            raise NotImplementedError(
                'The mechanistic model has no implemented set_parameter_names '
                'method.')

    def simulate(self, parameters, times):
        """
        Returns the numerical solution of the model outputs (and optionally
        the sensitivites) for the specified parameters and times.

        The model outputs are returned as a 2 dimensional NumPy array of shape
        ``(n_outputs, n_times)``. If sensitivities are enabled, a tuple is
        returned with the NumPy array of the model outputs and a NumPy array of
        the sensitivities of shape ``(n_times, n_outputs, n_parameters)``.

        :param parameters: An array-like object with values for the model
            parameters.
        :type parameters: list, numpy.ndarray
        :param times: An array-like object with time points at which the output
            values are returned.
        :type times: list, numpy.ndarray

        :rtype: np.ndarray of shape (n_outputs, n_times) or
            (n_times, n_outputs, n_parameters)
        """
        # Insert fixed parameter values
        if self._fixed_params_mask is not None:
            self._fixed_params_values[
                ~self._fixed_params_mask] = parameters
            parameters = self._fixed_params_values

        return self._mechanistic_model.simulate(parameters, times)

    def supports_dosing(self):
        """
        Returns a boolean whether dose administration with
        :meth:`PKPDModel.set_dosing_regimen` is supported by the model.
        """
        return self._mechanistic_model.supports_dosing()

    def time_unit(self):
        """
        Returns the model's unit of time.
        """
        try:
            time_unit = self._mechanistic_model.time_unit()
        except AttributeError:
            raise NotImplementedError(
                'The mechanistic model has no implemented time_unit '
                'method.')

        return time_unit



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import copy

import pints
import plotly.colors
import plotly.graph_objects as go
import xarray as xr

from chi import plots


class MarginalPosteriorPlot(plots.MultiSubplotFigure):
    """
    A figure class that visualises the marginal posterior probability for each
    parameter across individuals.

    One figure is generated for each parameter, which contains a marginal
    histogram of the sampled parameter values for each individual. The
    estimates for each indiviudal are plotted next to each other.

    This figure can be used to assess the convergence of the sampling method,
    as well as the variation of parameter estimates across individuals.

    Extends :class:`MultiFigure`.
    """
    def __init__(self):
        super(MarginalPosteriorPlot, self).__init__()

    def _add_histogram_plots(self, fig_id, data, colors):
        """
        Adds histogram plots of the parameter samples for each individual to
        the figure.

        One figure contains only the histograms of one parameter.
        """
        # Get number of colours
        n_colors = len(colors)

        # Check whether data are samples of a population parameter or
        # contains samples for many individuals
        is_population = False
        try:
            # Get ids of individuals
            ids = data.individual.values
        except AttributeError:
            is_population = True

        if is_population is True:
            # Compute diagnostics
            samples = data.values
            diagnostics = self._compute_diagnostics(samples)

            # Add trace
            color = colors[0]
            index = 0
            individual = 'Population'
            samples = samples.flatten()
            self._add_trace(
                fig_id, index, individual, samples, diagnostics, color)

            return None

        # Add trace for each individual
        for index, individual in enumerate(ids):
            # Get data for individual
            samples = data.sel(individual=individual).dropna(dim='draw')

            # Compute diagnostics
            samples = samples.values
            diagnostics = self._compute_diagnostics(samples)

            # Add trace
            samples = samples.flatten()
            color = colors[index % n_colors]
            self._add_trace(
                fig_id, index, individual, samples, diagnostics, color)

    def _add_trace(
            self, fig_id, index, individual, samples, diagnostics, color):
        """
        Adds a histogram of an indiviudals samples to a figure.
        """
        # Get figure
        fig = self._figs[fig_id]

        # Add trace
        rhat, = diagnostics
        fig.add_trace(
            go.Histogram(
                y=samples,
                name='%s' % str(individual),
                hovertemplate=(
                    'Sample: %{y:.2f}<br>' +
                    'Rhat: %.02f<br>' % rhat),
                visible=True,
                marker=dict(color=color),
                opacity=0.8),
            row=1,
            col=index+1)

        # Turn off xaxis ticks
        fig.update_xaxes(
            tickvals=[],
            row=1,
            col=index+1)

        # Set x axis title to individual
        fig.update_xaxes(
            title_text=str(individual), row=1, col=index+1)

    def _compute_diagnostics(self, data):
        """
        Computes and returns convergence metrics.

        - Rhat
        """
        # Compute rhat
        rhat = pints.rhat(chains=data)

        # Collect diagnostics
        diagnostics = [rhat]

        return diagnostics

    def add_data(self, data):
        """
        Adds marginal histograms of the samples across runs to the figure.

        The histograms of population parameters are visualised in separate
        figures,  while the individual parameters for one parameter type
        are grouped together.

        :param data: A :class:`xarray.Dataset` with the posterior samples.
        :type data: xarray.Dataset
        """
        # Check input format
        if not isinstance(data, xr.Dataset):
            raise TypeError(
                'The data has to be a xarray.Dataset.')

        dims = sorted(list(data.dims))
        expected_dims = ['chain', 'draw', 'individual']
        if (len(dims) == 2):
            expected_dims = ['chain', 'draw']
        for dim in expected_dims:
            if dim not in dims:
                raise ValueError(
                    'The data must have the dimensions '
                    '(chain, draw, individual). The current dimensions are <'
                    + str(dims) + '>.')

        # Get a colours
        colors = plotly.colors.qualitative.Plotly

        # Create a template figure (assigns it to self._fig)
        try:
            n_ids = len(data.individual)
        except AttributeError:
            # If data does not have individual attribute, all parameters
            # are population parameters
            n_ids = 1
        self._create_template_figure(
            rows=1, cols=n_ids, x_title='Normalised counts', spacing=0.01)

        # Create one figure for each parameter
        figs = []
        parameters = list(data.data_vars.keys())
        for parameter in parameters:
            # Check that parameter has as many ids as columns in template
            # figure
            try:
                number_ids = len(data[parameter].individual)
            except AttributeError:
                # If parameter does not have individual attribute, it's a
                # population parameter
                number_ids = 1

            if number_ids != n_ids:
                # Overwrite old n_ids
                n_ids = number_ids

                # Create a new template
                self._create_template_figure(
                    rows=1, cols=n_ids, x_title='Normalised counts',
                    spacing=0.01)

            # Append a copy of the template figure to all figures
            figs.append(copy.copy(self._fig))

        self._figs = figs

        # Add samples to parameter figures
        for index, parameter in enumerate(parameters):
            # Set y label of plot to parameter name
            self._figs[index].update_yaxes(
                title_text=parameter, row=1, col=1)

            # Get estimates for this parameter
            samples = data[parameter]

            # Add marginal histograms for all individuals
            self._add_histogram_plots(index, samples, colors)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import numpy as np
import pandas as pd
import plotly.colors
import plotly.graph_objects as go

from chi import plots


class ResidualPlot(plots.SingleFigure):
    """
    A figure class that visualises the residual error between the predictions
    of a predictive model and measured observations.

    Expects a :class:`pandas.DataFrame` of measurements with an ID, a time,
    observable and a value column. This dataset is used as reference to
    compute the residuals.

    Extends :class:`SingleFigure`.

    Parameters
    ----------
    data
        A :class:`pandas.DataFrame` with the time series PD data in form of
        an ID, time, and biomarker column.
    id_key
        Key label of the :class:`DataFrame` which specifies the ID column.
        The ID refers to the identity of an individual. Defaults to
        ``'ID'``.
    time_key
        Key label of the :class:`DataFrame` which specifies the time
        column. Defaults to ``'Time'``.
    obs_key
        Key label of the :class:`DataFrame` which specifies the observable
        column. Defaults to ``'Observable'``.
    value_key
        Key label of the :class:`DataFrame` which specifies the column of
        the measured values. Defaults to ``'Value'``.
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(
            self, measurements, id_key='ID', time_key='Time',
            obs_key='Observable', value_key='Value', updatemenu=True):
        super(ResidualPlot, self).__init__(updatemenu)

        # Check input format
        if not isinstance(measurements, pd.DataFrame):
            raise TypeError(
                'Measurements has to be pandas.DataFrame.')

        for key in [id_key, time_key, obs_key, value_key]:
            if key not in measurements.keys():
                raise ValueError(
                    'Measurements does not have the key <' + str(key) + '>.')

        # Remember data and keys
        self._measurements = measurements
        self._keys = [id_key, time_key, obs_key, value_key]

    def _add_predicted_versus_observed_scatter_plot(
            self, meas, pred, show_residuals, show_relative, time_key,
            sample_key):
        """
        Adds a scatter plot of the mean predictions on the x-axis and
        the measured values on the y-axis. Each individual gets a
        different colour.
        """
        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly
        n_colors = len(colors)

        # Get measurement keys
        id_key, time_key_m, _, meas_key = self._keys

        # Add scatter plot for each individual
        ids = meas[id_key].unique()
        for index, _id in enumerate(ids):
            # Get relevant measurements
            mask = meas[id_key] == _id
            temp = meas[mask]
            times = temp[time_key_m]
            observations = temp[meas_key].to_numpy()
            mean_predictions = self._get_mean_predictions(
                pred, times, time_key, sample_key)

            if show_residuals is True:
                # Compute residuals of observations from mean predictions
                observations -= mean_predictions

            if show_relative is True:
                # Normalise observations by mean predictions
                observations /= mean_predictions

            # Plot mean predictions versus observations
            color = colors[index % n_colors]
            self._add_residual_trace(
                _id, mean_predictions, observations, color)

        # Add default axes labels
        xlabel = 'Prediction'
        ylabel = 'Residual' if show_residuals is True else 'Observable'
        if show_relative is True:
            ylabel += ' in rel. units'
        self._fig.update_layout(
            xaxis_title=xlabel,
            yaxis_title=ylabel)

    def _add_residual_trace(
            self, _id, mean_predictions, measurements, color):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=mean_predictions,
                y=measurements,
                name="ID: %d" % _id,
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))))

    def _get_mean_predictions(
            self, pred, times, time_key, sample_key):
        """
        Returns a list of mean prediction estimates for the provided times.
        """
        means = np.empty(shape=len(times))
        for time_id, time in enumerate(times):
            # Compute mean
            mask = pred[time_key] == time
            means[time_id] = pred[mask][sample_key].mean()

        return means

    def _get_relevant_measurements(
            self, data, biomarker, individual, time_key):
        """
        Filters the observations for the relevant biomarker and ID. Also makes
        sure that there is a prediction for each measured time point.
        """
        # Get keys of measurement dataframe
        id_key_m, time_key_m, biom_key_m, _ = self._keys

        # Mask measurements for individual (if None keep all individuals)
        measurements = self._measurements
        if individual is not None:
            mask = measurements[id_key_m] == individual
            measurements = measurements[mask]

        # Mask measurements for biomarker
        mask = measurements[biom_key_m] == biomarker
        measurements = measurements[mask]

        # Make sure that there are predictions for each observed time
        measured_times = measurements[time_key_m].dropna().unique()
        predicted_times = data[time_key].to_numpy()
        for time in measured_times:
            if time not in predicted_times:
                raise ValueError(
                    'The prediction dataframe is not compatible with the '
                    'measurement dataframe. The prediction dataframe does not '
                    'provide predictions for the measurement time <%.3f>'
                    % time)

        return measurements

    def add_data(
            self, data, observable=None, individual=None, show_residuals=True,
            show_relative=False, time_key='Time', obs_key='Observable',
            value_key='Value'):
        r"""
        Adds the residuals of the predicted values with respect
        to the measured values to the figure.

        Expects a :class:`pandas.DataFrame` with a time, an observable and a
        value column. The time column determines the times of the
        measurements and the value column the measured
        values. The observable column determines the measured observable.

        The predictions are matched to the observations based on their ID and
        time label. If multiple predictions are provided for one measured time
        point, the mean prediction is computed as reference.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD simulation in
            form of a time and observable column.
        observable
            The predicted bimoarker. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            type in the observable column is selected.
        individual
            The ID of the individual whose measurements are used as reference
            for the predictive residuals. Defaults to ``None`` which compares
            the predictions to all individuals.
        show_residuals
            A boolean flag which indicates whether the residuals are plotted
            on the y axis, or the measurements themselves. Defaults to
            ``True``.
        show_relative
            A boolean flag which indicates whether the observations/residuals
            are normalised by the mean predictions. Defaults to ``False``.
        time_key
            Key label of the :class:`pandas.DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            value column. Defaults to ``'Value'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        for key in [time_key, obs_key, value_key]:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Check that selected individual exists in measurement dataframe
        if individual is not None:
            id_key_m = self._keys[0]
            ids = list(self._measurements[id_key_m].unique())
            if individual not in ids:
                raise ValueError(
                    'The ID <' + str(individual) + '> does not exist in the '
                    'measurement dataframe.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Check that selected observable exists in the measurement dataframe
        obs_key_m = self._keys[2]
        observables = self._measurements[obs_key_m].unique()
        if observable not in observables:
            raise ValueError(
                'The observable <' + str(observable) + '> does not exist in '
                'the measurement dataframe.')

        # Mask predictions for observable
        mask = data[obs_key] == observable
        data = data[mask]

        # Get the relevant observations
        meas = self._get_relevant_measurements(
            data, observable, individual, time_key)

        # Add mean predictions versus observations as scatter points
        self._add_predicted_versus_observed_scatter_plot(
            meas, data, show_residuals, show_relative, time_key,
            value_key)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

from ._covariate_models import (  # noqa
    CovariateModel,
    LinearCovariateModel
)

from ._error_models import (  # noqa
    ConstantAndMultiplicativeGaussianErrorModel,
    ErrorModel,
    GaussianErrorModel,
    LogNormalErrorModel,
    MultiplicativeGaussianErrorModel,
    ReducedErrorModel
)

from ._log_pdfs import (  # noqa
    HierarchicalLogLikelihood,
    HierarchicalLogPosterior,
    LogLikelihood,
    LogPosterior,
    PopulationFilterLogPosterior
)

from ._mechanistic_models import (  # noqa
    MechanisticModel,
    SBMLModel,
    PKPDModel,
    ReducedMechanisticModel
)

from ._inference import (  # noqa
    compute_pointwise_loglikelihood,
    InferenceController,
    OptimisationController,
    SamplingController
)

from ._population_filters import (  # noqa
    PopulationFilter,
    ComposedPopulationFilter,
    GaussianFilter,
    GaussianKDEFilter,
    GaussianMixtureFilter,
    LogNormalFilter,
    LogNormalKDEFilter
)

from ._population_models import (  # noqa
    ComposedPopulationModel,
    CovariatePopulationModel,
    GaussianModel,
    HeterogeneousModel,
    LogNormalModel,
    PooledModel,
    PopulationModel,
    ReducedPopulationModel,
    TruncatedGaussianModel
)

from ._predictive_models import (  # noqa
    AveragedPredictiveModel,
    PosteriorPredictiveModel,
    PredictiveModel,
    PopulationPredictiveModel,
    PriorPredictiveModel,
    PAMPredictiveModel
)

from ._problems import (  # noqa
    ProblemModellingController
)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

from ._data_library_api import (  # noqa
    DataLibrary
)

from ._model_library_api import (  # noqa
    ModelLibrary
)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import os

import pandas as pd


class DataLibrary(object):
    r"""
    A collection of Erlotinib PKPD datasets.

    Each method corresponds to a separate dataset, which will return
    the corresponding dataset in form of a :class:`pandas.DataFrame`.

    All dataset are organised in 9 columns:
    ID | Time | Time unit | Observable | Value | Observable unit | Dose |
    Dose unit | Duration.

    References
    ----------
    .. [1] Eigenmann, M. J. et al., Combining Nonclinical Experiments with
        Translational PKPD Modeling to Differentiate Erlotinib and
        Gefitinib, Mol Cancer Ther. 2016; 15(12):3110-3119.
    """

    def __init__(self):
        # Get path to data library
        self._path = os.path.dirname(os.path.abspath(__file__))
        self._path += '/data_library/'

    def lung_cancer_control_group(self):
        r"""
        Returns the lung cancer control group data published in [1]_ as a
        :class:`pandas.DataFrame`.

        The dataset contains the time series data of 8 mice with
        patient-derived lung cancer implants. The tumour volume of each
        mouse was monitored over a period of 30 days and measured a couple
        times a week.
        """
        file_name = 'lxf_control_growth.csv'
        data = pd.read_csv(self._path + file_name)

        return data

    def lung_cancer_high_erlotinib_dose_group(self):
        r"""
        Returns the high erlotinib dose lung cancer treatment group data
        published in [1]_ as a :class:`pandas.DataFrame`.

        The dataset contains the time series data of 6 mice with
        patient-derived lung cancer implants. Each mouse was treated with
        an oral dose of erlotinib of :math:`100\, \text{mg}` per
        :math:`\text{g}` body weight. The dose was administered daily from
        day 3 to day 16, with a treatment break on days 9 to 13.

        The blood plasma concentration of erlotinib was measured on day 14,
        while the tumour volume of each mouse was monitored over a period
        of 30 days and measured a couple times a week.
        """
        file_name = 'lxf_high_erlotinib_dose.csv'
        data = pd.read_csv(self._path + file_name)

        return data

    def lung_cancer_low_erlotinib_dose_group(self):
        r"""
        Returns the low erlotinib dose lung cancer treatment group data
        published in [1]_ as a :class:`pandas.DataFrame`.

        The dataset contains the time series data of 8 mice with
        patient-derived lung cancer implants. Each mouse was treated with
        an oral dose of erlotinib of :math:`6.25\, \text{mg}` per
        :math:`\text{g}` body weight. The dose was administered daily from
        day 3 to day 16.

        The blood plasma concentration of erlotinib was measured on day 10 and
        16, while the tumour volume of each mouse was monitored over a period
        of 30 days and measured a couple times a week.
        """
        file_name = 'lxf_low_erlotinib_dose.csv'
        data = pd.read_csv(self._path + file_name)

        return data

    def lung_cancer_medium_erlotinib_dose_group(self):
        r"""
        Returns the medium erlotinib dose lung cancer treatment group data
        published in [1]_ as a :class:`pandas.DataFrame`.

        The dataset contains the time series data of 8 mice with
        patient-derived lung cancer implants. Each mouse was treated with
        an oral dose of erlotinib of :math:`25\, \text{mg}` per
        :math:`\text{g}` body weight. The dose was administered daily from
        day 3 to day 16.

        The blood plasma concentration of erlotinib was measured on day 10 and
        16, while the tumour volume of each mouse was monitored over a period
        of 30 days and measured a couple times a week.
        """
        file_name = 'lxf_medium_erlotinib_dose.csv'
        data = pd.read_csv(self._path + file_name)

        return data

    def lung_cancer_single_erlotinib_dose_group(self):
        r"""
        Returns the single erlotinib dose lung cancer treatment group data
        published in [1]_ as a :class:`pandas.DataFrame`.

        The dataset contains the time series data of 30 mice with
        patient-derived lung cancer implants. Each mouse was treated with
        a single oral dose of erlotinib of :math:`100\, \text{mg}` per
        :math:`\text{g}` body weight. The dose was administered either on
        day 0 or day 4.

        The blood plasma concentration of erlotinib was measured only once per
        mouse, either on day 0 or day 4.
        """
        file_name = 'lxf_single_erlotinib_dose.csv'
        data = pd.read_csv(self._path + file_name)

        return data



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import os

import chi


class ModelLibrary(object):
    """
    Contains references to pharmacokinetic and pharmacodynamic models in SBML
    file format.

    These models can be instantiated for simulation or inference with a
    :class:`MechanisticModel`.

    References
    ----------
    .. [1] Koch, G. et al. Modeling of tumor growth and anticancer effects
        of combination therapy. J Pharmacokinet Pharmacodyn 36, 179–197
        (2009).
    """

    def __init__(self):
        # Get path to model library
        self._path = os.path.dirname(os.path.abspath(__file__))
        self._path += '/model_library/'

    def erlotinib_tumour_growth_inhibition_model(self):
        """
        This model is a combination of a
        :meth:`ModelLibrary.one_compartment_pk_model` and a
        :meth:`tumour_growth_inhibition_model_koch_reparametrised`.
        """
        file_name = 'temporary_full_pkpd_model.xml'

        return chi.PKPDModel(self._path + file_name)

    def one_compartment_pk_model(self):
        r"""
        Returns an instantiation of a 1-compartment PK model.

        In this model the distribution of the drug is modelled by one
        compartment with a linear elimination rate :math:`k_e`

        .. math ::
            \frac{\text{d}A}{\text{d}t} = -k_e A \quad C = \frac{A}{V}.

        Here, :math:`A` and :math:`C` are the amount and the concentration of
        the drug in the body, respectively. :math:`V` is the effective volume
        of distribution of the drug in the compartment.

        This model may be interpreted as modelling the blood plasma
        concentration of the drug, with the assumption that the clearance of
        the drug through the liver may be approximated by an exponential decay
        with the rate :math:`k_e`.

        The drug may be either directly administered to :math:`A` or indirectly
        through a dosing compartment.

        :rtype: chi.PKPDModel
        """
        file_name = 'pk_one_comp.xml'
        model = chi.PKPDModel(self._path + file_name)
        model.set_outputs(['central.drug_concentration'])

        return model

    def tumour_growth_inhibition_model_koch(self):
        r"""
        Returns the absolute path to a SBML file, specifying the tumour growth
        inhibition pharmacodynamic model introduced by Koch et al. in [1]_.

        In this model the tumour growth inhibition is modelled by an empirical
        model of the tumour volume :math:`V_T` over time

        .. math::
            \frac{\text{d}V_T}{\text{d}t} =
                \frac{2\lambda_0\lambda_1 V_T}
                {2\lambda_0V_T + \lambda_1} - \kappa C V_T.


        Here, the tumour growth in absence of the drug is assumed to grow
        exponentially at rate :math:`2\lambda_0` for tumour volumes below some
        critical volume :math:`V_{\text{crit}}`. For volumes beyond
        :math:`V_{\text{crit}}` the growth dynamics is assumed to slow down
        and transition to a linear growth at rate :math:`\lambda_0`. The tumour
        growth inhibitory effect of the compound is modelled proportionally to
        its concentration :math:`C` and the current tumour volume. The
        proportionality factor :math:`\kappa` can be interpreted as the potency
        of the drug.

        Note that the critical tumour volume :math:`V_{\text{crit}}` at which
        the growth dynamics transitions from exponential to linear growth is
        given by the two growth rates

        .. math::
            V_{\text{crit}} = \frac{\lambda _1}{2\lambda _0}.

        :rtype: chi.SBMLModel
        """
        file_name = 'tgi_Koch_2009.xml'

        return chi.SBMLModel(self._path + file_name)

    def tumour_growth_inhibition_model_koch_reparametrised(self):
        r"""
        Returns the absolute path to a SBML file, specifying the tumour growth
        inhibition pharmacodynamic model introduced by Koch et al. in [1]_ with
        modified parametrisation.

        In this model the tumour growth inhibition is modelled by an empirical
        model of the tumour volume :math:`V_T` over time

        .. math::
            \frac{\text{d}V_T}{\text{d}t} =
                \frac{\lambda V_T}
                {V_T / V_{\text{crit}} + 1} - \kappa C V_T.


        Here, the tumour growth in absence of the drug is assumed to grow
        exponentially at rate :math:`\lambda` for tumour volumes below some
        critical volume :math:`V_{\text{crit}}`. For volumes beyond
        :math:`V_{\text{crit}}` the growth dynamics is assumed to slow down
        and transition to a linear growth at rate
        :math:`\lambda V_{\text{crit}}`. The tumour growth inhibitory effect
        of the compound is modelled proportionally to its concentration
        :math:`C` and the current tumour volume. The proportionality factor
        :math:`\kappa` can be interpreted as the potency of the drug.

        Note that this parameterisation of the model is related to the original
        parametersation in [1]_ by

        .. math::
            V_{\text{crit}} = \frac{\lambda _1}{2\lambda _0} \quad \text{and}
            \quad \lambda = 2\lambda _1 .

        :rtype: chi.SBMLModel
        """
        file_name = 'tgi_Koch_2009_reparametrised.xml'

        return chi.SBMLModel(self._path + file_name)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import numpy as np
import pandas as pd
import plotly.colors
import plotly.graph_objects as go

from chi import plots


class PDPredictivePlot(plots.SingleFigure):
    """
    A figure class that visualises the predictions of a predictive
    pharmacodynamic model.

    Extends :class:`SingleFigure`.

    Parameters
    ----------
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(self, updatemenu=True):
        super(PDPredictivePlot, self).__init__(updatemenu)

    def _add_data_trace(self, _id, times, measurements, color):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=measurements,
                name="ID: %d" % _id,
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))))

    def _add_prediction_scatter_trace(self, times, samples):
        """
        Adds scatter plot of samples from the predictive model.
        """
        # Get colour (light blueish)
        color = plotly.colors.qualitative.Pastel2[1]

        # Add trace
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=samples,
                name="Predicted samples",
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))))

    def _add_prediction_bulk_prob_trace(self, data):
        """
        Adds the bulk probabilities as two line plots (one for upper and lower
        limit) and shaded area to the figure.
        """
        # Construct times that go from min to max and back to min
        # (Important for shading with 'toself')
        times = data['Time'].unique()
        times = np.hstack([times, times[::-1]])

        # Get unique bulk probabilities and sort in descending order
        bulk_probs = data['Bulk probability'].unique()
        bulk_probs[::-1].sort()

        # Get colors (shift start a little bit, because 0th level is too light)
        n_traces = len(bulk_probs)
        shift = 2
        colors = plotly.colors.sequential.Blues[shift:shift+n_traces]

        # Add traces
        for trace_id, bulk_prob in enumerate(bulk_probs):
            # Get relevant upper and lower percentiles
            mask = data['Bulk probability'] == bulk_prob
            reduced_data = data[mask]

            upper = reduced_data['Upper'].to_numpy()
            lower = reduced_data['Lower'].to_numpy()
            values = np.hstack([upper, lower[::-1]])

            # Add trace
            self._fig.add_trace(go.Scatter(
                x=times,
                y=values,
                line=dict(width=1, color=colors[trace_id]),
                fill='toself',
                legendgroup='Model prediction',
                name='Predictive model',
                text="%s Bulk" % bulk_prob,
                hoverinfo='text',
                showlegend=True if trace_id == n_traces-1 else False))

    def _compute_bulk_probs(self, data, bulk_probs, time_key, sample_key):
        """
        Computes the upper and lower percentiles from the predictive model
        samples, corresponding to the provided bulk probabilities.
        """
        # Create container for perecentiles
        container = pd.DataFrame(columns=[
            'Time', 'Upper', 'Lower', 'Bulk probability'])

        # Translate bulk probabilities into percentiles
        percentiles = []
        for bulk_prob in bulk_probs:
            lower = 0.5 - bulk_prob / 2
            upper = 0.5 + bulk_prob / 2

            percentiles.append([bulk_prob, lower, upper])

        # Get unique times
        unique_times = data[time_key].unique()

        # Fill container with percentiles for each time
        for time in unique_times:
            # Mask relevant data
            mask = data[time_key] == time
            reduced_data = data[mask]

            # Get percentiles
            percentile_df = reduced_data[sample_key].rank(
                pct=True)
            for item in percentiles:
                bulk_prob, lower, upper = item

                # Get biomarker value corresponding to percentiles
                mask = percentile_df <= lower
                biom_lower = reduced_data[mask][sample_key].max()

                mask = percentile_df >= upper
                biom_upper = reduced_data[mask][sample_key].min()

                # Append percentiles to container
                container = pd.concat([container, pd.DataFrame({
                    'Time': [time],
                    'Lower': [biom_lower],
                    'Upper': [biom_upper],
                    'Bulk probability': [str(bulk_prob)]})])

        return container

    def add_data(
            self, data, observable=None, id_key='ID', time_key='Time',
            obs_key='Observable', value_key='Value'):
        """
        Adds pharmacodynamic time series data of (multiple) individuals to
        the figure.

        Expects a :class:`pandas.DataFrame` with an ID, a time, an
        observable and a value column, and adds a scatter plot of the
        measured time series to the figure. Each individual receives a
        unique colour.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD data in form of
            an ID, time, and observable column.
        observable
            The predicted observable. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        id_key
            Key label of the :class:`DataFrame` which specifies the ID column.
            The ID refers to the identity of an individual. Defaults to
            ``'ID'``.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the column of
            the measured values. Defaults to ``'Value'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        for key in [id_key, time_key, obs_key, value_key]:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask]

        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly
        n_colors = len(colors)

        # Fill figure with scatter plots of individual data
        ids = data[id_key].unique()
        for index, _id in enumerate(ids):
            # Get individual data
            mask = data[id_key] == _id
            times = data[time_key][mask]
            measurements = data[value_key][mask]
            color = colors[index % n_colors]

            # Create Scatter plot
            self._add_data_trace(_id, times, measurements, color)

    def add_prediction(
            self, data, observable=None, bulk_probs=[0.9], time_key='Time',
            obs_key='Observable', value_key='Value'):
        r"""
        Adds the prediction to the figure.

        Expects a :class:`pandas.DataFrame` with a time, an observable and a
        value column. The time column determines the times of the
        measurements and the value column the measured value.
        The observable column determines the observable.

        A list of bulk probabilities ``bulk_probs`` can be specified, which are
        then added as area to the figure. The corresponding upper and lower
        percentiles are estimated from the ranks of the provided
        samples.

        .. warning::
            For low sample sizes the illustrated bulk probabilities may deviate
            significantly from the theoretical bulk probabilities. The upper
            and lower limit are determined from the rank of the samples for
            each time point.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD simulation in
            form of a time and observable column.
        observable
            The predicted observable. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        bulk_probs
            A list of bulk probabilities that are illustrated in the
            figure. If ``None`` the samples are illustrated as a scatter plot.
        time_key
            Key label of the :class:`pandas.DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            value column. Defaults to ``'Value'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        for key in [time_key, obs_key, value_key]:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask]

        # Add samples as scatter plot if no bulk probabilites are provided, and
        # terminate method
        if bulk_probs is None:
            times = data[time_key]
            samples = data[value_key]
            self._add_prediction_scatter_trace(times, samples)

            return None

        # Not more than 7 bulk probabilities are allowed (Purely aesthetic
        # criterion)
        if len(bulk_probs) > 7:
            raise ValueError(
                'At most 7 different bulk probabilities can be illustrated at '
                'the same time.')

        # Make sure that bulk probabilities are between 0 and 1
        bulk_probs = [float(probability) for probability in bulk_probs]
        for probability in bulk_probs:
            if (probability < 0) or (probability > 1):
                raise ValueError(
                    'The provided bulk probabilities have to between 0 and 1.')

        # Add bulk probabilities to figure
        percentile_df = self._compute_bulk_probs(
            data, bulk_probs, time_key, value_key)
        self._add_prediction_bulk_prob_trace(percentile_df)


class PKPredictivePlot(plots.SingleSubplotFigure):
    """
    A figure class that visualises the predictions of a predictive
    pharmacokinetic model.

    Extends :class:`SingleSubplotFigure`.

    Parameters
    ----------
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(self, updatemenu=True):
        super(PKPredictivePlot, self).__init__()

        self._create_template_figure(
            rows=2, cols=1, shared_x=True, row_heights=[0.2, 0.8])

        # Define legend name of prediction
        self._prediction_name = 'Predictive model'

        if updatemenu:
            self._add_updatemenu()

    def _add_dose_trace(
            self, _id, times, doses, durations, color,
            is_prediction=False):
        """
        Adds scatter plot of dose events to figure.
        """
        # Convert durations to strings
        durations = [
            'Dose duration: ' + str(duration) for duration in durations]

        name = "ID: %s" % str(_id)
        if is_prediction is True:
            name = 'Predictive model'

        # Add scatter plot of dose events
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=doses,
                name=name,
                legendgroup=name,
                showlegend=False,
                mode="markers",
                text=durations,
                hoverinfo='text',
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))),
            row=1,
            col=1)

    def _add_biom_trace(self, _id, times, measurements, color):
        """
        Adds scatter plot of an indiviudals pharamcokinetics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=measurements,
                name="ID: %s" % str(_id),
                legendgroup="ID: %s" % str(_id),
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))),
            row=2,
            col=1)

    def _add_updatemenu(self):
        """
        Adds a button to the figure that switches the biomarker scale from
        linear to logarithmic.
        """
        self._fig.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="left",
                    buttons=list([
                        dict(
                            args=[{"yaxis2.type": "linear"}],
                            label="Linear y-scale",
                            method="relayout"
                        ),
                        dict(
                            args=[{"yaxis2.type": "log"}],
                            label="Log y-scale",
                            method="relayout"
                        )
                    ]),
                    pad={"r": 0, "t": -10},
                    showactive=True,
                    x=0.0,
                    xanchor="left",
                    y=1.15,
                    yanchor="top"
                )
            ]
        )

    def _add_prediction_scatter_trace(self, times, samples):
        """
        Adds scatter plot of samples from the predictive model.
        """
        # Get colour (light blueish)
        color = plotly.colors.qualitative.Pastel2[1]

        # Add trace
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=samples,
                name="Predicted samples",
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))))

    def _add_prediction_bulk_prob_trace(self, data, colors):
        """
        Adds the bulk probabilities as two line plots (one for upper and lower
        limit) and shaded area to the figure.
        """
        # Construct times that go from min to max and back to min
        # (Important for shading with 'toself')
        times = data['Time'].unique()
        times = np.hstack([times, times[::-1]])

        # Get unique bulk probabilities and sort in descending order
        bulk_probs = data['Bulk probability'].unique()
        bulk_probs[::-1].sort()

        # Add traces
        n_traces = len(bulk_probs)
        for trace_id, bulk_prob in enumerate(bulk_probs):
            # Get relevant upper and lower percentiles
            mask = data['Bulk probability'] == bulk_prob
            reduced_data = data[mask]

            upper = reduced_data['Upper'].to_numpy()
            lower = reduced_data['Lower'].to_numpy()
            values = np.hstack([upper, lower[::-1]])

            # Add trace
            self._fig.add_trace(go.Scatter(
                x=times,
                y=values,
                line=dict(width=1, color=colors[trace_id]),
                fill='toself',
                legendgroup=self._prediction_name,
                name=self._prediction_name,
                text="%s Bulk" % bulk_prob,
                hoverinfo='text',
                showlegend=True if trace_id == n_traces-1 else False),
                row=2,
                col=1)

    def _compute_bulk_probs(self, data, bulk_probs, time_key, sample_key):
        """
        Computes the upper and lower percentiles from the predictive model
        samples, corresponding to the provided bulk probabilities.
        """
        # Create container for perecentiles
        container = pd.DataFrame(columns=[
            'Time', 'Upper', 'Lower', 'Bulk probability'])

        # Translate bulk probabilities into percentiles
        percentiles = []
        for bulk_prob in bulk_probs:
            lower = 0.5 - bulk_prob / 2
            upper = 0.5 + bulk_prob / 2

            percentiles.append([bulk_prob, lower, upper])

        # Get unique times
        unique_times = data[time_key].unique()

        # Fill container with percentiles for each time
        for time in unique_times:
            # Mask relevant data
            mask = data[time_key] == time
            reduced_data = data[mask]

            # Get percentiles
            percentile_df = reduced_data[sample_key].rank(
                pct=True)
            for item in percentiles:
                bulk_prob, lower, upper = item

                # Get biomarker value corresponding to percentiles
                mask = percentile_df <= lower
                biom_lower = reduced_data[mask][sample_key].max()

                mask = percentile_df >= upper
                biom_upper = reduced_data[mask][sample_key].min()

                # Append percentiles to container
                container = pd.concat([container, pd.DataFrame({
                    'Time': [time],
                    'Lower': [biom_lower],
                    'Upper': [biom_upper],
                    'Bulk probability': [str(bulk_prob)]})])

        return container

    def add_data(
            self, data, observable=None, id_key='ID', time_key='Time',
            obs_key='Observable', value_key='Value', dose_key='Dose',
            dose_duration_key='Duration'):
        """
        Adds pharmacokinetic time series data of (multiple) individuals to
        the figure.

        Expects a :class:`pandas.DataFrame` with an ID, a time, an
        observable and a value column, and adds a scatter plot of the
        measuremed time series to the figure. The dataframe is also expected
        to have information about the administered dose via a dose and a
        dose duration column. Each individual receives a unique colour.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD data in form of
            an ID, time, and observable column.
        observable
            The measured observable. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        id_key
            Key label of the :class:`DataFrame` which specifies the ID column.
            The ID refers to the identity of an individual. Defaults to
            ``'ID'``.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the column of
            the measured values. Defaults to ``'Value'``.
        dose_key
            Key label of the :class:`DataFrame` which specifies the dose
            column. Defaults to ``'Dose'``.
        dose_duration_key
            Key label of the :class:`DataFrame` which specifies the dose
            duration column. Defaults to ``'Duration'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        keys = [
            id_key, time_key, obs_key, value_key, dose_key, dose_duration_key]
        for key in keys:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Get dose information
        mask = data[dose_key].notnull()
        dose_data = data[mask][[id_key, time_key, dose_key, dose_duration_key]]

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask][[id_key, time_key, value_key]]

        # Set axis labels to dataframe keys
        self.set_axis_labels(time_key, obs_key, dose_key)

        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly
        n_colors = len(colors)

        # Fill figure with scatter plots of individual data
        ids = data[id_key].unique()
        for index, _id in enumerate(ids):
            # Get doses applied to individual
            mask = dose_data[id_key] == _id
            dose_times = dose_data[time_key][mask]
            doses = dose_data[dose_key][mask]
            durations = dose_data[dose_duration_key][mask]

            # Get observable measurements
            mask = data[id_key] == _id
            times = data[time_key][mask]
            measurements = data[value_key][mask]

            # Get a color for the individual
            color = colors[index % n_colors]

            # Create scatter plot of dose events
            self._add_dose_trace(_id, dose_times, doses, durations, color)

            # Create Scatter plot
            self._add_biom_trace(_id, times, measurements, color)

    def add_prediction(
            self, data, observable=None, bulk_probs=[0.9], time_key='Time',
            obs_key='Observable', value_key='Value', dose_key='Dose',
            dose_duration_key='Duration'):
        r"""
        Adds the prediction for the observable pharmacokinetic observable
        values to the figure.

        Expects a :class:`pandas.DataFrame` with a time, an observable and a
        value column. The time column determines the time of the observable
        measurement and the sample column the corresponding observable
        measurement. The observable column determines the observable type. The
        dataframe is also expected to have information about the administered
        dose via a dose and a dose duration column.

        A list of bulk probabilities ``bulk_probs`` can be specified, which are
        then added as area to the figure. The corresponding upper and lower
        percentiles are estimated from the ranks of the provided
        samples.

        .. warning::
            For low sample sizes the illustrated bulk probabilities may deviate
            significantly from the theoretical bulk probabilities. The upper
            and lower limit are determined from the rank of the samples for
            each time point.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD simulation in
            form of a time and observable column.
        observable
            The predicted observable. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        bulk_probs
            A list of bulk probabilities that are illustrated in the
            figure. If ``None`` the samples are illustrated as a scatter plot.
        time_key
            Key label of the :class:`pandas.DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`pandas.DataFrame` which specifies the
            value column. Defaults to ``'Value'``.
        dose_key
            Key label of the :class:`DataFrame` which specifies the dose
            column. Defaults to ``'Dose'``.
        dose_duration_key
            Key label of the :class:`DataFrame` which specifies the dose
            duration column. Defaults to ``'Duration'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        keys = [
            time_key, obs_key, value_key, dose_key, dose_duration_key]
        for key in keys:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Get dose information
        mask = data[dose_key].notnull()
        dose_data = data[mask][[time_key, dose_key, dose_duration_key]]

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask][[time_key, value_key]]

        # Set axis labels to dataframe keys
        self.set_axis_labels(time_key, obs_key, dose_key)

        # Add samples as scatter plot if no bulk probabilites are provided, and
        # terminate method
        if bulk_probs is None:
            times = data[time_key]
            samples = data[value_key]
            self._add_prediction_scatter_trace(times, samples)

            return None

        # Not more than 7 bulk probabilities are allowed (Purely aesthetic
        # criterion)
        if len(bulk_probs) > 7:
            raise ValueError(
                'At most 7 different bulk probabilities can be illustrated at '
                'the same time.')

        # Make sure that bulk probabilities are between 0 and 1
        bulk_probs = [float(probability) for probability in bulk_probs]
        for probability in bulk_probs:
            if (probability < 0) or (probability > 1):
                raise ValueError(
                    'The provided bulk probabilities have to between 0 and 1.')

        # Define colour scheme
        shift = 2
        colors = plotly.colors.sequential.Blues[shift:]

        # Create scatter plot of dose events
        self._add_dose_trace(
            _id=None,
            times=dose_data[time_key],
            doses=dose_data[dose_key],
            durations=dose_data[dose_duration_key],
            color=colors[0],
            is_prediction=True)

        # Add bulk probabilities to figure
        percentile_df = self._compute_bulk_probs(
            data, bulk_probs, time_key, value_key)
        self._add_prediction_bulk_prob_trace(percentile_df, colors)

    def set_axis_labels(self, time_label, biom_label, dose_label):
        """
        Sets the label of the time axis, the biomarker axis, and the dose axis.
        """
        self._fig.update_xaxes(title=time_label, row=2)
        self._fig.update_yaxes(title=dose_label, row=1)
        self._fig.update_yaxes(title=biom_label, row=2)


class PDTimeSeriesPlot(plots.SingleFigure):
    """
    A figure class that visualises measurements of a pharmacodynamic
    observables across multiple individuals.

    Measurements of a pharmacodynamic observables over time are visualised as a
    scatter plot.

    Extends :class:`SingleFigure`.

    Parameters
    ----------
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(self, updatemenu=True):
        super(PDTimeSeriesPlot, self).__init__(updatemenu)

    def _add_data_trace(self, _id, times, measurements, color):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=measurements,
                name="ID: %d" % _id,
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))))

    def _add_simulation_trace(self, times, biomarker):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=biomarker,
                name="Model",
                showlegend=True,
                mode="lines",
                line=dict(color='black')))

    def add_data(
            self, data, observable=None, id_key='ID', time_key='Time',
            obs_key='Observable', value_key='Value'):
        """
        Adds pharmacodynamic time series data of (multiple) individuals to
        the figure.

        Expects a :class:`pandas.DataFrame` with an ID, a time, an
        observable and a value column, and adds a scatter plot of the
        measuremed time series to the figure. Each individual receives a
        unique colour.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD data in form of
            an ID, time, and observable column.
        observable
            The measured bimoarker. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        id_key
            Key label of the :class:`DataFrame` which specifies the ID column.
            The ID refers to the identity of an individual. Defaults to
            ``'ID'``.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the column of
            the measured values. Defaults to ``'Value'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        for key in [id_key, time_key, obs_key, value_key]:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask]

        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly
        n_colors = len(colors)

        # Fill figure with scatter plots of individual data
        ids = data[id_key].unique()
        for index, _id in enumerate(ids):
            # Get individual data
            mask = data[id_key] == _id
            times = data[time_key][mask]
            measurements = data[value_key][mask]
            color = colors[index % n_colors]

            # Create Scatter plot
            self._add_data_trace(_id, times, measurements, color)

    def add_simulation(self, data, time_key='Time', value_key='Value'):
        """
        Adds a pharmacodynamic time series simulation to the figure.

        Expects a :class:`pandas.DataFrame` with a time and a value
        column, and adds a line plot of the simulated time series to the
        figure.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD simulation in
            form of a time and value column.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the
            value column. Defaults to ``'Value'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        for key in [time_key, value_key]:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        times = data[time_key]
        values = data[value_key]

        self._add_simulation_trace(times, values)


class PKTimeSeriesPlot(plots.SingleSubplotFigure):
    """
    A figure class that visualises measurements of a pharmacokinetic observable
    across multiple individuals.

    Measurements of a pharmacokinetic observable over time are visualised as a
    scatter plot.

    Extends :class:`SingleSubplotFigure`.

    Parameters
    ----------
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(self, updatemenu=True):
        super(PKTimeSeriesPlot, self).__init__()

        self._create_template_figure(
            rows=2, cols=1, shared_x=True, row_heights=[0.2, 0.8])

        if updatemenu:
            self._add_updatemenu()

    def _add_dose_trace(self, _id, times, doses, durations, color):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        # Convert durations to strings
        durations = [
            'Dose duration: ' + str(duration) for duration in durations]

        # Add scatter plot of dose events
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=doses,
                name="ID: %s" % str(_id),
                legendgroup="ID: %s" % str(_id),
                showlegend=False,
                mode="markers",
                text=durations,
                hoverinfo='text',
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))),
            row=1,
            col=1)

    def _add_biom_trace(self, _id, times, measurements, color):
        """
        Adds scatter plot of an indiviudals pharamcodynamics to figure.
        """
        self._fig.add_trace(
            go.Scatter(
                x=times,
                y=measurements,
                name="ID: %s" % str(_id),
                legendgroup="ID: %s" % str(_id),
                showlegend=True,
                mode="markers",
                marker=dict(
                    symbol='circle',
                    color=color,
                    opacity=0.7,
                    line=dict(color='black', width=1))),
            row=2,
            col=1)

    def _add_updatemenu(self):
        """
        Adds a button to the figure that switches the biomarker scale from
        linear to logarithmic.
        """
        self._fig.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="left",
                    buttons=list([
                        dict(
                            args=[{"yaxis2.type": "linear"}],
                            label="Linear y-scale",
                            method="relayout"
                        ),
                        dict(
                            args=[{"yaxis2.type": "log"}],
                            label="Log y-scale",
                            method="relayout"
                        )
                    ]),
                    pad={"r": 0, "t": -10},
                    showactive=True,
                    x=0.0,
                    xanchor="left",
                    y=1.15,
                    yanchor="top"
                )
            ]
        )

    def add_data(
            self, data, observable=None, id_key='ID', time_key='Time',
            obs_key='Observable', value_key='Value', dose_key='Dose',
            dose_duration_key='Duration'):
        """
        Adds pharmacokinetic time series data of (multiple) individuals to
        the figure.

        Expects a :class:`pandas.DataFrame` with an ID, a time, an
        observable and a value column, and adds a scatter plot of the
        measuremed time series to the figure. The dataframe is also expected
        to have information about the administered dose via a dose and a
        dose duration column. Each individual receives a unique colour.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD data in form of
            an ID, time, observable and value column.
        observable
            The measured bimoarker. This argument is used to determine the
            relevant rows in the dataframe. If ``None``, the first observable
            in the observable column is selected.
        id_key
            Key label of the :class:`DataFrame` which specifies the ID column.
            The ID refers to the identity of an individual. Defaults to
            ``'ID'``.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        obs_key
            Key label of the :class:`DataFrame` which specifies the
            observable column. Defaults to ``'Observable'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the column of
            the measured values. Defaults to ``'Value'``.
        dose_key
            Key label of the :class:`DataFrame` which specifies the dose
            column. Defaults to ``'Dose'``.
        dose_duration_key
            Key label of the :class:`DataFrame` which specifies the dose
            duration column. Defaults to ``'Duration'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        keys = [
            id_key, time_key, obs_key, value_key, dose_key, dose_duration_key]
        for key in keys:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Default to first bimoarker, if observable is not specified
        biom_types = data[obs_key].dropna().unique()
        if observable is None:
            observable = biom_types[0]

        if observable not in biom_types:
            raise ValueError(
                'The observable could not be found in the observable column.')

        # Get dose information
        mask = data[dose_key].notnull()
        dose_data = data[mask][[id_key, time_key, dose_key, dose_duration_key]]

        # Mask data for observable
        mask = data[obs_key] == observable
        data = data[mask][[id_key, time_key, value_key]]

        # Set axis labels to dataframe keys
        self.set_axis_labels(time_key, obs_key, dose_key)

        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly
        n_colors = len(colors)

        # Fill figure with scatter plots of individual data
        ids = data[id_key].unique()
        for index, _id in enumerate(ids):
            # Get doses applied to individual
            mask = dose_data[id_key] == _id
            dose_times = dose_data[time_key][mask]
            doses = dose_data[dose_key][mask]
            durations = dose_data[dose_duration_key][mask]

            # Get observable measurements
            mask = data[id_key] == _id
            times = data[time_key][mask]
            measurements = data[value_key][mask]

            # Get a color for the individual
            color = colors[index % n_colors]

            # Create scatter plot of dose events
            self._add_dose_trace(_id, dose_times, doses, durations, color)

            # Create Scatter plot
            self._add_biom_trace(_id, times, measurements, color)

    def add_simulation(
            self, data, time_key='Time', value_key='Value',
            dose_key='Dose'):
        """
        Adds a pharmacokinetic time series simulation to the figure.

        Expects a :class:`pandas.DataFrame` with a time, a value,
        and a dose column. A line plot of the biomarker time series, as well
        as the dosing regimen is added to the figure.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the time series PD simulation in
            form of a time and a value column.
        time_key
            Key label of the :class:`DataFrame` which specifies the time
            column. Defaults to ``'Time'``.
        value_key
            Key label of the :class:`DataFrame` which specifies the simulated
            values column. Defaults to ``'Value'``.
        """
        raise NotImplementedError

    def set_axis_labels(self, time_label, biom_label, dose_label):
        """
        Sets the label of the time axis, the biomarker axis, and the dose axis.
        """
        self._fig.update_xaxes(title=time_label, row=2)
        self._fig.update_yaxes(title=dose_label, row=1)
        self._fig.update_yaxes(title=biom_label, row=2)



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import plotly.graph_objects as go
from plotly.subplots import make_subplots


class SingleFigure(object):
    """
    Base class for plot classes that generate a single figure.

    Parameters
    ----------
    updatemenu
        Boolean flag that enables or disables interactive buttons, such as a
        logarithmic scale switch for the y-axis.
    """

    def __init__(self, updatemenu=True):
        super(SingleFigure, self).__init__()

        self._fig = go.Figure()
        self._set_layout(updatemenu)

    def _set_layout(self, updatemenu):
        """
        Configures the basic layout of the figure.

        - Size
        - Template
        - Log and linear y scale switches
        """
        self._fig.update_layout(
            autosize=True,
            template="plotly_white")

        if updatemenu:
            self._fig.update_layout(
                updatemenus=[
                    dict(
                        type="buttons",
                        direction="left",
                        buttons=list([
                            dict(
                                args=[{"yaxis.type": "linear"}],
                                label="Linear y-scale",
                                method="relayout"
                            ),
                            dict(
                                args=[{"yaxis.type": "log"}],
                                label="Log y-scale",
                                method="relayout"
                            )
                        ]),
                        pad={"r": 0, "t": -10},
                        showactive=True,
                        x=0.0,
                        xanchor="left",
                        y=1.15,
                        yanchor="top"
                    )
                ]
            )

    def add_data(self, data):
        """
        Adds data to the figure.
        """
        raise NotImplementedError

    def set_axis_labels(self, xlabel, ylabel):
        """
        Sets the x axis, and y axis label of the figure.
        """
        self._fig.update_layout(
            xaxis_title=xlabel,
            yaxis_title=ylabel)

    def show(self):
        """
        Displays the figure.
        """
        self._fig.show()


class SingleSubplotFigure(object):
    """
    Base class for plot classes that a single figure with subplots.
    """

    def __init__(self):
        super(SingleSubplotFigure, self).__init__()

        self._fig = go.Figure()
        self._set_layout()

    def _create_template_figure(
            self, rows, cols, x_title=None, y_title=None, shared_x=False,
            shared_y=False, spacing=0.05, column_widths=None,
            row_heights=None):
        """
        Creates a template figure using :meth:`plotly.make_subplots`.
        """
        self._fig = make_subplots(
            rows=rows, cols=cols, shared_xaxes=shared_x, shared_yaxes=shared_y,
            x_title=x_title, y_title=y_title, horizontal_spacing=spacing,
            vertical_spacing=spacing, column_widths=column_widths,
            row_heights=row_heights)

        # Set layout
        self._set_layout()

    def _set_layout(self):
        """
        Configures the basic layout of the figure.

        - Size
        - Template
        """
        self._fig.update_layout(
            autosize=True,
            template="plotly_white")

    def show(self):
        """
        Displays the figure.
        """
        self._fig.show()


class MultiFigure(object):
    """
    Base class for plot classes that generate multiple figures.
    """

    def __init__(self, updatemenu=True):
        super(MultiFigure, self).__init__()

        # Create a template figure
        self._fig = go.Figure()
        self._set_layout(updatemenu)
        self._figs = [self._fig]

    def _set_layout(self, updatemenu):
        """
        Configures the basic layout of the figure.

        - Size
        - Template
        - Log and linear y scale switches
        """
        self._fig.update_layout(
            autosize=True,
            template="plotly_white")

        if updatemenu:
            self._fig.update_layout(
                updatemenus=[
                    dict(
                        type="buttons",
                        direction="left",
                        buttons=list([
                            dict(
                                args=[{"yaxis.type": "linear"}],
                                label="Linear y-scale",
                                method="relayout"
                            ),
                            dict(
                                args=[{"yaxis.type": "log"}],
                                label="Log y-scale",
                                method="relayout"
                            )
                        ]),
                        pad={"r": 0, "t": -10},
                        showactive=True,
                        x=0.0,
                        xanchor="left",
                        y=1.15,
                        yanchor="top"
                    )
                ]
            )

    def add_data(self, data):
        """
        Adds data to the figure.
        """
        raise NotImplementedError

    def show(self):
        """
        Displays the figures.
        """
        for fig in self._figs:
            fig.show()


class MultiSubplotFigure(MultiFigure):
    """
    Base class for plot classes that generate multiple figures with subplots.

    Extends :class:`MultiFigure`.
    """

    def __init__(self):
        super(MultiSubplotFigure, self).__init__(updatemenu=False)

    def _create_template_figure(
            self, rows, cols, x_title=None, y_title=None, shared_x=False,
            shared_y=True, spacing=0.05):
        """
        Creates a template figure using :meth:`plotly.make_subplots`.
        """
        self._fig = make_subplots(
            rows=rows, cols=cols, shared_xaxes=shared_x, shared_yaxes=shared_y,
            x_title=x_title, y_title=y_title, horizontal_spacing=spacing,
            vertical_spacing=spacing)

        # Set layout
        self._set_layout(updatemenu=False)
